#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): Cody J. Balos @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

"""
This script can be used to parse Caliper data if Caliper was run with the
hatchet-region-profile config. The expected format is json-split.
See http://software.llnl.gov/Caliper/BuiltinConfigurations.html and
https://hatchet.readthedocs.io/en/latest/analysis_examples.html#json-file.

The script requires numpy, pandas and hatchet:
  pip install hatchet
  pip install pandas
  pip install numpy

It can also use [Flamegraph](https://github.com/brendangregg/FlameGraph) to
generate a nice stack-trace-like plot, if Flamegraph is available in your path.
"""

import argparse
import hatchet as ht
import numpy as np
import pandas as pd
import subprocess
import tempfile

def run_flamegraph(txt_file, fg_args):
    """
    Call Flamegraph.
    """
    subprocess.run(["flamegraph.pl"] + fg_args + [txt_file])


def Flamegraph(gf, fg_args):
    with tempfile.NamedTemporaryFile() as folded_stack:
        folded_stack.write(str.encode(gf.to_flamegraph()))
        folded_stack.seek(0)
        run_flamegraph(folded_stack.name, fg_args)


parser = argparse.ArgumentParser(description='Utility for parsing/plotting Caliper data with Hatchet.')
subparsers = parser.add_subparsers(help='sub-command help', dest='command')

fg_parser = subparsers.add_parser('flamegraph')
fg_parser.add_argument('--fg-bin',
                       help='path to Flamegraph binary')
fg_parser.add_argument('json_file',
                       help='path to the Caliper json-split data file (generated when the hatchet-region-profile Caliper config is enabled)')
fg_parser.add_argument('fg_args', help='Flamegraph arguments', nargs=argparse.REMAINDER)

ht_parser = subparsers.add_parser('hatchet')
ht_parser.add_argument('json_file',
                       help='path to the Caliper json-split data file (generated when the hatchet-region-profile Caliper config is enabled)')

args = parser.parse_args()
# print(args)

# Use hatchet's ``from_caliper_json`` API with the resulting json-split.
# The result is stored into Hatchet's GraphFrame.
gf = ht.GraphFrame.from_caliper_json(args.json_file)

if args.command == 'flamegraph':
    Flamegraph(gf, args.fg_args)
elif args.command == 'hatchet':
    print(gf.tree())



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Function to parse SUNDIALS CSV output files
# -----------------------------------------------------------------------------

def num(s):
    """Try to convert a string to an int or float"""

    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def keys(filename):
    """Extracts keys from a SUNDIALS CSV file

    Parameters
    ----------
    filename : str
        The file location of the SUNDIALS CSV file

    Returns
    -------
    list
        A list of dictionary keys
    """

    # Get keys from the first row
    with open(filename, mode='r') as csvfile:
        keys = csvfile.readline().split(",")[::2]

    return keys


def read(filename):
    """Reads a SUNDIALS CSV file

    Parameters
    ----------
    filename : str
        The file location of the SUNDIALS CSV file

    Returns
    -------
    dict
        A dictionary containing the CSV keys and values
    """

    import csv

    # Get dictionary keys
    fields = keys(filename)

    # Initialize dictionary
    csv_dict = {}
    for k in fields:
        csv_dict[k] = []

    # Get values from each row
    with open(filename, mode='r') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            values = row[1::2]
            for k, v in zip(fields, values):
                csv_dict[k].append(num(v))

    return csv_dict


def write(filename):
    """Prints a SUNDIALS CSV file

    Parameters
    ----------
    filename : str
        The file location of the SUNDIALS CSV file
    """

    csv_dict = read(filename)

    for key in csv_dict.keys():
        print(f"{key:29} = {csv_dict[key]}")
    print()



#!/usr/bin/env python
# ----------------------------------------------------------------
# Programmer(s):  Daniel R. Reynolds @ SMU
# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
# matplotlib-based plotting script for ODE examples

# imports
import sys
import pylab as plt
import numpy as np

# load solution data file
data = np.loadtxt('solution.txt', dtype=np.double)

# determine number of time steps, number of fields
nt,nv = np.shape(data)

# extract time array
times = data[:,0]

# parse comment line to determine solution names
f = open('solution.txt', 'r')
commentline = f.readline()
commentsplit = commentline.split()
names = commentsplit[2:]

# create plot
plt.figure()

# add curves to figure
for i in range(nv-1):
    plt.plot(times,data[:,i+1],label=names[i])
plt.xlabel('t')
if (nv > 2):
    plt.ylabel('solutions')
else:
    plt.ylabel('solution')
plt.legend(loc='upper right', shadow=True)
plt.grid()
plt.savefig('solution.png')




##### end of script #####



#! /usr/bin/env python
# -------------------------------------------------------------------------------
# Programmer(s): Eddy Banks and David J. Gardner @ LLNL
# -------------------------------------------------------------------------------
# This is a python script for running nightly tests on the most recent commit to
# the SUNDIALS git repository. Results are stored in /usr/casc/sundials/devtest
# and the main log file is sent to the SUNDIALS developers mailing list
# "sundials-devs@llnl.gov" after all tests are completed.
# -------------------------------------------------------------------------------

import sys, os, shutil
import subprocess
import datetime, time
import smtplib
from email.MIMEText import MIMEText
from optparse import OptionParser

def main():

    maxCommitDirs = 5               # max number of commit directories to keep
    maxBuildsPerPreviousCommit = 3  # max number of builds to keep per previous commit
    maxBuildsPerCurrentCommit  = 10 # max number of builds to keep for current commit

    sunGitRepo = "ssh://git@mystash.llnl.gov:7999/sundials/sunrepo.git"

    sunBaseDir = "/usr/casc/sundials/devtest"
    sunNightlyDir = os.path.join(sunBaseDir,"nightly")

    # main log file for this script
    buildLogFileName = "build.log"
    buildLogFile = os.path.join(sunNightlyDir, buildLogFileName)
    sys.stdout = Logger(buildLogFile)

    # other log files created in this script
    cmakeLogFileName = "cmake.log"
    makeLogFileName  = "make.log"

    # log file created by the testRunner script
    testLogFileName  = "build/Testing/Temporary/LastTest.log"

    # place holders in case an error is encountered before these variables are created
    makeLogFile  = "Oops - didn't get this far!"
    cmakeLogFile = "Oops - didn't get this far!"
    testLogFile  = "Oops - didn't get this far!"
    sunTestDir   = "Oops - didn't get this far!"
    msg          = "Oops - didn't get this far!"

    todayDate = time.strftime("%Y-%m-%d")

    print "*** SUNDIALS Automated Build/Test ***"
    startTime = datetime.datetime.now()
    print "\nStart time:", startTime.ctime()

    try:
        # get most recent commit hash
        cmd = "git ls-remote " + sunGitRepo + " | grep HEAD"
        cmdout = runCommand(cmd)

        commitHashLong  = cmdout.split()[0].strip()
        commitHashShort = commitHashLong[0:7]
        print "Git commit:",commitHashLong

        # create commit test directory
        sunTestDir = os.path.join(sunNightlyDir,
                                  "commit_"+commitHashShort,
                                  todayDate)
        print "\n*** Creating commit test directory: " + sunTestDir
        if not os.path.exists(sunTestDir):
            os.makedirs(sunTestDir)

        # change working directory to test dir
        os.chdir(sunTestDir)

        # if commit test directory is not empty - exit
        if os.listdir(sunTestDir) != []:
            msg = "'" + sunTestDir + "' is not empty."
            raise Exception(msg)

        # checkout only the most recent commit to the repo
        cmd = "git clone --quiet --depth 1 " + sunGitRepo
        print "\n*** Cloning most recent commit: " + cmd
        cmdout = runCommand(cmd)

        # change working directory to clone of sundials repo
        sunSrcDir = os.path.join(sunTestDir,"sunrepo")
        os.chdir(sunSrcDir)

        # print most recent commit message
        cmd = "git log -1"
        print "\n*** Most recent commit message: " + cmd + " \n"
        cmdout = runCommand(cmd)
        print cmdout

        # change working directory to test dir
        os.chdir(sunTestDir)

        # create build directory
        sunBuildDir = os.path.join(sunTestDir, "build")
        print "\n*** Creating build directory: " + sunBuildDir
        if not os.path.exists(sunBuildDir):
            os.makedirs(sunBuildDir)

        # if build directory is not empty - exit
        if os.listdir(sunBuildDir) != []:
            msg = "'" + sunBuildDir + "' is not empty."
            raise Exception(msg)

        # change working directory to build dir
        os.chdir(sunBuildDir)

        # run CMake to configure
        cmakeLogFile = os.path.join(sunBuildDir, cmakeLogFileName)
        cmd = "cmake \ \n"
        # set compiler flags to check for non-standard code
        # -ansi  OR  -std-c89  OR  -std=c99  OR  -std=c11
        # NOTE: PETSC requires -std=c99 or newer
        cmd = cmd + "-DCMAKE_C_FLAGS='-Wall -std=c99 -pedantic' \ \n"
        # enable mpi
        cmd = cmd + "-DENABLE_MPI=ON \ \n"
        # enable C++
        cmd = cmd + "-DCXX_ENABLE=TRUE \ \n"
        # enable lapack   (NOTE: will find libraries in LD_LIBRARY_PATH)
        cmd = cmd + "-DENABLE_LAPACK=ON \ \n"
        # enable klu
        cmd = cmd + "-DENABLE_KLU=ON \ \n"
        cmd = cmd + "-DKLU_INCLUDE_DIR=/usr/casc/sundials/apps/rh6/suitesparse/4.5.3/include \ \n"
        cmd = cmd + "-DKLU_LIBRARY_DIR=/usr/casc/sundials/apps/rh6/suitesparse/4.5.3/lib \ \n"
        # enable hypre
        cmd = cmd + "-DENABLE_HYPRE=ON \ \n"
        cmd = cmd + "-DHYPRE_INCLUDE_DIR=/usr/casc/sundials/apps/rh6/hypre/2.11.1_long_int_fpic/include \ \n"
        cmd = cmd + "-DHYPRE_LIBRARY_DIR=/usr/casc/sundials/apps/rh6/hypre/2.11.1_long_int_fpic/lib \ \n"
        # enable PETSc
        cmd = cmd + "-DENABLE_PETSC=ON \ \n"
        cmd = cmd + "-DPETSC_INCLUDE_DIR=/usr/casc/sundials/apps/rh6/petsc/3.7.2_long_int/include \ \n"
        cmd = cmd + "-DPETSC_LIBRARY_DIR=/usr/casc/sundials/apps/rh6/petsc/3.7.2_long_int/lib \ \n"
        # enable openmp
        cmd = cmd + "-DENABLE_OPENMP=ON \ \n"
        # enable pthreads
        cmd = cmd + "-DENABLE_PTHREAD=ON \ \n"
        # enable SUPERLU_MT
        cmd = cmd + "-DENABLE_SUPERLUMT=ON \ \n"
        cmd = cmd + "-DSUPERLUMT_INCLUDE_DIR=/usr/casc/sundials/apps/rh6/superlu_mt/SuperLU_MT_3.1_long_int_fpic/SRC \ \n"
        cmd = cmd + "-DSUPERLUMT_LIBRARY_DIR=/usr/casc/sundials/apps/rh6/superlu_mt/SuperLU_MT_3.1_long_int_fpic/lib \ \n"
        cmd = cmd + "-DSUPERLUMT_THREAD_TYPE=Pthread \ \n"
        # turn on development tests
        cmd = cmd + "-DSUNDIALS_TEST_DEVTESTS=ON \ \n"
        # specify source
        cmd = cmd + sunSrcDir
        # redirect output to config log file
        cmd = cmd + " &> " + cmakeLogFile

        print "\n*** Running CMake:\n" + cmd

        # remove newlines from cmd
        cmd = cmd.replace('\n','')
        cmd = cmd.replace('\\','')
        cmdout = runCommand(cmd)
        print cmdout

        # run make to build libs and executables
        makeLogFile = os.path.join(sunBuildDir, makeLogFileName)
        cmd = "make -j4 &> " + makeLogFile
        print "\n*** Building with: " + cmd
        cmdout = runCommand(cmd)
        print cmdout

        # set location of test runner log file
        testLogFile = os.path.join(sunBuildDir, testLogFileName)

        # run 'make test' to test examples
        cmd = "make test"
        print "\n*** Testing with: " + cmd
        cmdout = runCommand(cmd)
        print cmdout

        # find the percentage of successful tests
        percentLine = findPercentSuccessfulLine(cmdout)

        # tests finished
        msg = "Finished: Commit: " + commitHashShort + " | " + percentLine

        # purge nightly build directory (keep maxRevDirs number of revisions)
        print "\n*** Purging nightly build directory"
        commitDirList = purgeNightlyDir(sunNightlyDir, maxCommitDirs)

        # now purge each revision directory (except most current)
        for commitDir in commitDirList[:-1]:
            purgeNightlyDir(commitDir, maxBuildsPerPreviousCommit)

        # now purge current revision directory (last entry in list)
        purgeNightlyDir(commitDirList[-1], maxBuildsPerCurrentCommit)

    except Exception as e:
        msg = "FAILED: Commit: " + commitHashShort + " | " + str(e)
        return 1

    finally:
        # clean up test after success or fail
        cleanup(msg, startTime, sunTestDir, buildLogFile, buildLogFileName,
                cmakeLogFile, makeLogFile, testLogFile)

    return 0

# ===============================================================================

#
# class to redirect system output to log file and screen
#
class Logger(object):
    def __init__(self, logFile):
        self.terminal = sys.stdout
        self.log = open(logFile, "w")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        self.log.flush()

    def close(self):
        self.log.close()

#
# run external command
#
def runCommand(cmd):
    cmdout = subprocess.check_output(cmd, shell=True)

    return(cmdout)

#
# find % test successful line in 'make test' output
#
def findPercentSuccessfulLine(cmdout):
    # search for line with first word containing %
    percentLine = ""
    for line in cmdout.splitlines():
        if '%' in line.split(' ')[0]:
            percentLine = line
            break

    return percentLine

#
# sort directories by modification time
#
def sortDirByTime(rootDir):
    mtime   = lambda x: os.stat(os.path.join(rootDir, x)).st_mtime
    dirList = sorted(os.listdir(rootDir), key=mtime)
    return dirList

#
# purge the nightly directory of older builds
#
def purgeNightlyDir(rootDir, maxKeepDirs):
    dirList = []
    for tmpName in sortDirByTime(rootDir):
        tmpSpec = os.path.join(rootDir, tmpName)
        if os.path.isdir(tmpSpec):
            dirExtension = os.path.splitext(tmpSpec)[1]
            # only purge directories without extensions
            # (e.g. allows user to add '.save' to specific build)
            if dirExtension == '':
                dirList.append(tmpSpec)
                #print tmpSpec

    # if more than maxRevDirs - remove them
    numDirs = len(dirList)
    removeDirCount = numDirs - maxKeepDirs
    for i in xrange(removeDirCount):
        shutil.rmtree(dirList[i])

    # return list of remaining directories
    return dirList[removeDirCount:numDirs]

#
# cleanup after test
#
def cleanup(msg, startTime, sunTestDir, buildLogFile, buildLogFileName,
            cmakeLogFile, makeLogFile, testLogFile):

    # move log file to test directory
    if (os.path.isdir(sunTestDir)):
        finalBuildLogFile = os.path.join(sunTestDir, buildLogFileName)
        os.rename(buildLogFile, finalBuildLogFile)
    else:
        finalBuildLogFile = buildLogFile

    # print closing info
    print "\n" + msg
    print "For build script details see:    " + finalBuildLogFile
    print "For CMake specific details see:  " + cmakeLogFile
    print "For make specific details see:   " + makeLogFile
    print "For test details see:            " + testLogFile

    endTime = datetime.datetime.now()
    print "\nEnd time:", endTime.ctime()
    elapsedTime = endTime - startTime
    print "Elapsed Time:", elapsedTime

    sys.stdout.flush()
    sys.stdout.close()

    # send email
    sendEmail(finalBuildLogFile, msg)

    return

#
# send email to SUNDIALS dev list
#
def sendEmail(logFile, subject):
    # Open a plain text file for reading.  For this example, assume that
    # the text file contains only ASCII characters.
    fp = open(logFile, 'rb')

    # Create a text/plain message
    msg = MIMEText(fp.read())
    fp.close()

    ### me == the sender's email address
    me = "SUNDIALS.sunbuild@llnl.gov"
    ### you == the recipient's email address
    you = "sundials-devs@llnl.gov"
    msg['Subject'] = subject
    msg['From'] = me
    msg['To'] = you

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    s = smtplib.SMTP('smtp.llnl.gov')
    s.sendmail(me, [you], msg.as_string())
    s.quit()

# ===============================================================================

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# ------------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# ------------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ------------------------------------------------------------------------------
# Find unique lines from a file containing a specific string and write them to
# a separate file.
#
# git diff --numstat | awk '{ added += $1; removed += $2 } END { print "+" added " -" removed }'
#
# Example usage:
#   Find unique compiler warnings in make.log and write them to warnings.txt:
#   $ ./findlines.py make.log warning warnings.txt
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# main routine
# ------------------------------------------------------------------------------
def main():

    import argparse
    import sys, os
    import shlex
    from subprocess import call

    parser = argparse.ArgumentParser(
        description='Count the number lines changed in a certain path',
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('key1',type=str,
                        help='First git identifier')

    parser.add_argument('key2',type=str,
                        help='Second git identifier')

    parser.add_argument('patterns',type=str,nargs='+',
                        help='File path patterns to look for in diff')


    # parse command line args
    args = parser.parse_args()

    # git stats from git
    cmd = 'git diff --numstat '+args.key1+' '+args.key2+' > tmp.txt'
    print cmd
    call(cmd, shell=True)

    totaladded   = 0
    totalremoved = 0
    netchange    = 0

    with open('tmp.txt',"r") as fr:
        for line in fr:
            split_line = shlex.split(line)
            added    = split_line[0]
            removed  = split_line[1]
            filename = split_line[2]
            if (any(x in filename for x in args.patterns)):
                if ((added == '-') or (removed == '-')):
                    continue
                else:
                    print added, removed, filename
                    totaladded   += int(added)
                    totalremoved += int(removed)

    netchange = totaladded - totalremoved
    print totaladded, totalremoved, netchange

# ------------------------------------------------------------------------------
# run the main routine
# ------------------------------------------------------------------------------
if __name__ == '__main__':
    import sys
    sys.exit(main())



#!/usr/bin/env python
# ------------------------------------------------------------
# Programmer(s):  Daniel R. Reynolds @ SMU
# ------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ------------------------------------------------------------
# matplotlib-based plotting script for brusselator1D.c example

# imports
import sys
import pylab as plt
import numpy as np

# load mesh data file
mesh = np.loadtxt('bruss_FEM_mesh.txt', dtype=np.double)

# load solution data files
udata = np.loadtxt('bruss_FEM_u.txt', dtype=np.double)
vdata = np.loadtxt('bruss_FEM_v.txt', dtype=np.double)
wdata = np.loadtxt('bruss_FEM_w.txt', dtype=np.double)

# determine number of time steps, mesh size
nt,nx = np.shape(udata)

# determine min/max values
umin = 0.9*udata.min()
umax = 1.1*udata.max()
vmin = 0.9*vdata.min()
vmax = 1.1*vdata.max()
wmin = 0.9*wdata.min()
wmax = 1.1*wdata.max()
minval = np.array([umin, vmin, wmin]).min()
maxval = np.array([umax, vmax, wmax]).max()

# plot the mesh
plt.figure(1)
plt.plot(mesh,0.0*mesh,'o')
plt.xlabel('x')
plt.title('FEM mesh')
plt.savefig('brusselator1D_FEM_mesh.png')

# generate plots of results
for tstep in range(nt):

    # set string constants for output plots, current time, mesh size
    pname = 'brusselator1D_FEM.' + repr(tstep).zfill(3) + '.png'
    tstr  = repr(tstep)
    nxstr = repr(nx)

    # plot current solution and save to disk
    plt.figure(1)
    plt.plot(mesh,udata[tstep,:],label="u")
    plt.plot(mesh,vdata[tstep,:],label="v")
    plt.plot(mesh,wdata[tstep,:],label="w")
    plt.xlabel('x')
    plt.ylabel('solution')
    plt.title('Solutions at output ' + tstr + ', mesh = ' + nxstr)
    plt.axis((0.0, 1.0, minval, maxval))
    plt.grid()
    plt.legend(loc='upper right', shadow=True)
    plt.savefig(pname)
    plt.close()


##### end of script #####



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

# imports
import glob
import sys
import matplotlib
matplotlib.use('Agg')
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


# load pickled data
def load_data(file):
    data = np.load(file)
    m = data['mesh']
    t = data['t']
    u = data['u']
    v = data['v']
    w = data['w']

    hx = m[0,1] - m[0,0]
    hy = m[1,1] - m[1,0]
    hz = m[2,1] - m[2,0]

    return { 'm': m, 'h': (hx,hy,hz), 't': t, 'u': u, 'v': v, 'w': w }


# grid function norm
def norm_3Dgrid(h, x, q=1):
    hx,hy,hz = h
    s = np.shape(x)
    return (hx*hy*hz*np.sum(np.abs(x)**q, axis=(1,2,3)))**(1./q)


# load data files
np111 = load_data('np-111/output-with-h-8.33e-02.npz')
np211 = load_data('np-211/output-with-h-8.33e-02.npz')
np311 = load_data('np-311/output-with-h-8.33e-02.npz')
np131 = load_data('np-131/output-with-h-8.33e-02.npz')
np113 = load_data('np-113/output-with-h-8.33e-02.npz')
np911 = load_data('np-911/output-with-h-8.33e-02.npz')
# np133 = load_data('np-133/output-with-h-8.33e-02.npz')
np313 = load_data('np-313/output-with-h-8.33e-02.npz')
np331 = load_data('np-331/output-with-h-8.33e-02.npz')
np333 = load_data('np-333/output-with-h-8.33e-02.npz')
# np666 = load_data('np-666/output-with-h-8.33e-02.npz')

for component in ['u', 'v', 'w']:
    # Reference solution
    ref = np111[component]

    # Now compute E(h) = ||U(h) - \bar{U}(h)|| using the grid-function norm
    E_np211 = norm_3Dgrid(np211['h'], np211[component] - ref)
    E_np311 = norm_3Dgrid(np311['h'], np311[component] - ref)
    E_np131 = norm_3Dgrid(np131['h'], np131[component] - ref)
    E_np113 = norm_3Dgrid(np113['h'], np113[component] - ref)
    E_np911 = norm_3Dgrid(np911['h'], np911[component] - ref)
    # E_np133 = norm_3Dgrid(np133['h'], np133[component] - ref)
    E_np313 = norm_3Dgrid(np313['h'], np313[component] - ref)
    E_np331 = norm_3Dgrid(np331['h'], np331[component] - ref)
    E_np333 = norm_3Dgrid(np333['h'], np333[component] - ref)
    # E_np666 = norm_3Dgrid(np666['h'], np666[component] - ref)

    # Plot error across time
    X, Y = np.meshgrid(np111['m'][0,:], np111['t'])
    # fig = plt.figure()
    # ax = plt.subplot(311, projection='3d')
    # ax.plot_surface(X, Y, np.abs(np911[component][:,:,0,0] - ref[:,:,0,0]))
    # ax = plt.subplot(312, projection='3d')
    # ax.plot_surface(X, Y, np.abs(np911[component][:,0,:,0] - ref[:,0,:,0]))
    # ax = plt.subplot(313, projection='3d')
    # ax.plot_surface(X, Y, np.abs(np911[component][:,0,0,:] - ref[:,0,0,:]))
    plt.plot(np111['t'], E_np211)
    plt.plot(np111['t'], E_np131)
    plt.plot(np111['t'], E_np113)
    plt.plot(np111['t'], E_np911)
    # plt.plot(np111['t'], E_np133)
    plt.plot(np111['t'], E_np313)
    plt.plot(np111['t'], E_np331)
    plt.plot(np111['t'], E_np333)
    # plt.plot(np111['t'], E_np666)
    # plt.legend(['2 1 1', '3 1 1', '1 3 3', '3 1 3', '3 3 1', '3 3 3', '6 6 6'])
    # plt.legend(['3 1 1', '1 3 1', '1 1 3', '9 1 1', '1 3 3', '3 1 3', '3 3 1'])
    plt.ylabel('||E(hx,hy,hz)||')
    plt.xlabel('time')
    plt.savefig('compare-error-plot-%s.png' % component)



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Send email notification if a SUNDIALS regression test status
# -----------------------------------------------------------------------------

def main():

    import argparse
    import os

    parser = argparse.ArgumentParser(
        description='Send email notification based on regression test status',
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('teststatus', type=str,
                        choices=['passed', 'failed', 'fixed'],
                        help='Status of regression test')

    parser.add_argument('testname', type=str,
                        help='Name branch name or pull-request tested')

    parser.add_argument('testurl', type=str,
                        help='URL for viewing test results')

    # parse command line args
    args = parser.parse_args()

    # name of regression test log file
    logfile = "suntest.log"

    # if log file exists add url, otherwise create log file
    if (os.path.isfile(logfile)):
        with open(logfile, "a") as log:
            log.write("View test output at:\n")
            log.write(args.testurl)
    else:
        print("Warning: log file not found")
        with open(logfile, "w") as log:
            log.write("Warning: log file not found\n")
            log.write("View test output at:\n")
            log.write(args.testurl)

    # determine notification recipient
    special_branches = ['main', 'develop', 'release']

    if any(branch in args.testname for branch in special_branches):
        # SUNDIALS developers list
        recipient = "sundials-devs@llnl.gov"
    else:
        # author of most recent commit
        cmd = "git log --format='%ae' -1"
        recipient = runCommand(cmd).rstrip()

        # check if the last commit was a CI merge
        if (recipient == 'nobody@nowhere'):
            cmd = "git log HEAD~1 --pretty=format:'%ae' -1"
            recipient = runCommand(cmd).rstrip()

    # send notification if tests fail, log file not found, or fixed
    if (args.teststatus == 'failed'):

        subject = "FAILED: SUNDIALS "+args.testname+" failed regression tests"
        print("Tests failed, sending notification to", recipient)
        sendEmail(recipient, subject, logfile)

    elif (args.teststatus == 'fixed'):

        subject = "FIXED: SUNDIALS "+args.testname+" passed regression tests"
        print("Tests fixed, sending notification to", recipient)
        sendEmail(recipient, subject, logfile)

    else:
        print("Tests passed, no notifications sent")


#
# run external command
#
def runCommand(cmd):

    import subprocess

    cmdout = subprocess.check_output(cmd, shell=True)

    return(cmdout)


#
# send email
#
def sendEmail(recipient, subject, message):

    import smtplib
    from email.MIMEText import MIMEText

    # Open a plain text file for reading. Assumed that
    # the text file contains only ASCII characters.
    fp = open(message, 'rb')

    # Create a text/plain message
    msg = MIMEText(fp.read())
    fp.close()

    # sender's email address
    sender = "SUNDIALS.suntest@llnl.gov"

    # email settings
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipient

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    s = smtplib.SMTP('smtp.llnl.gov')
    s.sendmail(sender, [recipient], msg.as_string())
    s.quit()


#
# just run the main routine
#
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# This script plots the output from test_nector_performance_* and assumes:
#   1. vector lengths are powers of two starting from 0, and
#   2. output files are named: output_nelem_nvec_nsum_ntest_timing.txt
# where nelem is the number of elements in the vector, nvec is the nuber of
# vectors, nsum is the number of sums, ntest is the number of tests, and timing
# indicates if timing was enabled.
# -----------------------------------------------------------------------------

def main():

    import argparse
    import os, sys
    import shlex
    import glob

    import numpy as np
    import scipy.stats as st

    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.ticker as mtick

    parser = argparse.ArgumentParser(
        description='Plot data from NVector performance tests')

    parser.add_argument('op', type=str,
                        help='Which NVector operation to plot')

    parser.add_argument('datadir', type=str,
                        help='Directory where test output files are located')

    parser.add_argument('--timevelem', dest='timevelem', action='store_true',
                        help='Turn on plots for time vs number of elements')

    parser.add_argument('--noheatmap', dest='heatmap', action='store_false',
                        help='Turn off heatmap plots')

    parser.add_argument('--loglog', dest='loglog', action='store_true',
                        help='Generate loglog plots for time vs number of elements')

    parser.add_argument('--show', dest='show', action='store_true',
                        help='Display plots rather than saving to file')

    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Turn on debugging output')

    # parse command line args
    args = parser.parse_args()

    if (args.debug):
        print(args)

    # check for test data directory
    if (not os.path.isdir(args.datadir)):
        print("ERROR:",args.datadir,"does not exist")
        sys.exit()

    # sort output files
    output = sorted(glob.glob(args.datadir+'/output*.txt'))

    # if (args.debug):
    #     print("output files")
    #     print(len(output))
    #     for i in range(len(output)):
    #         print(output[i])

    # figure out vector sizes, number of vectors, and number of sums
    nelem = []
    nvec  = []
    nsum  = []
    ntest = []

    # parse file names to get input parameters
    for f in output:

        split_fout = f.split("/")[-1]
        split_fout = split_fout.split("_")

        ne = int(split_fout[1])
        nv = int(split_fout[2])
        ns = int(split_fout[3])
        nt = int(split_fout[4])

        if (not ne in nelem):
            nelem.append(ne)

        if (not nv in nvec):
            nvec.append(nv)

        if (not ns in nsum):
            nsum.append(ns)

        if (not nt in ntest):
            ntest.append(nt)

    if (len(ntest) != 1):
        print("Warning: Unequal numbers of tests")

    if (args.debug):
        print("nelem:",nelem, len(nelem))
        print("nvec: ",nvec,  len(nvec))
        print("nsum: ",nsum,  len(nsum))
        print("ntest:",ntest, len(ntest))

    # allocate numpy arrays for timing data
    avg_fused  = np.zeros([len(nvec), len(nelem)])
    sdev_fused = np.zeros([len(nvec), len(nelem)])

    avg_unfused  = np.zeros([len(nvec), len(nelem)])
    sdev_unfused = np.zeros([len(nvec), len(nelem)])

    avg_ratio = np.zeros([len(nvec), len(nelem)])

    # NVEC = np.zeros([len(nvec), len(nelem)])
    # NELM = np.zeros([len(nvec), len(nelem)])

    # read output files
    for f in output:

        if (args.debug):
            print("Reading:",f)

        # get test inputs from file name
        split_fout = f.split("/")[-1]
        split_fout = split_fout.split("_")

        ne = int(split_fout[1])
        nv = int(split_fout[2])
        ns = int(split_fout[3])

        with open(f) as fout:
            for line in fout:

                # split line into list
                split_line = shlex.split(line)

                # skip blank lines
                if (not split_line):
                    continue

                # tests finished, stop reading file
                if (split_line[0] == "Finished"):
                    break

                # check if the operation is the one we want and get data
                if (args.op == split_line[0]):

                    i = nvec.index(nv)
                    j = nelem.index(ne)

                    # NVEC[i][j] = nv
                    # NELM[i][j] = ne

                    avg_fused[i][j]  = float(split_line[1])
                    sdev_fused[i][j] = float(split_line[2])

                    avg_unfused[i][j]  = float(split_line[5])
                    sdev_unfused[i][j] = float(split_line[6])

                    avg_ratio[i][j] = avg_fused[i][j] / avg_unfused[i][j]

    if (args.debug):
        print(avg_fused)
        print(avg_unfused)
        print(avg_ratio)
        print(sdev_fused)
        print(sdev_unfused)

    # --------------------------------------------------------------------------
    # Confidence Interval
    # --------------------------------------------------------------------------

    # allocate arrays for the upper and lower bounds of the confidence interval
    lower_fused   = np.zeros([len(nvec), len(nelem)])
    upper_fused   = np.zeros([len(nvec), len(nelem)])
    lower_unfused = np.zeros([len(nvec), len(nelem)])
    upper_unfused = np.zeros([len(nvec), len(nelem)])

    # critical value for 99% confidence interval
    if (ntest[0] < 30):
        # student's t distribution
        cv = st.t.interval(0.99, ntest[0]-1)[1]
    else:
        # normal distribution
        cv = st.norm.ppf(0.995)

    # confidence intervals
    cdev_fused  = cv * sdev_fused / np.sqrt(ntest[0])
    lower_fused = avg_fused - cdev_fused
    upper_fused = avg_fused + cdev_fused

    cdev_unfused  = cv * sdev_unfused / np.sqrt(ntest[0])
    lower_unfused = avg_unfused - cdev_unfused
    upper_unfused = avg_unfused + cdev_unfused

    # check if the fused average times are within the unfused confidence interval
    fused_in = np.where(np.logical_and(avg_fused < upper_unfused,
                                       avg_fused > lower_unfused))

    # check if the unfused average times are within the fused confidence interval
    unfused_in = np.where(np.logical_and(avg_unfused < upper_fused,
                                         avg_unfused > lower_fused))

    # get which numbers of vectors and elements for fused tests are in the
    # confidence interval of the unfused times
    df = np.zeros([len(fused_in[0])])
    vf = np.zeros([len(fused_in[0])])
    ef = np.zeros([len(fused_in[0])])

    for i in range(len(fused_in[0])):
        vf[i] = nvec[fused_in[0][i]]
        ef[i] = np.log2(nelem[fused_in[1][i]])
        df[i] = 1

    if (args.debug):
        print(vf)
        print(ef)

    # get which numbers of vectors and elements for unfused tests are in the
    # confidence interval of the fused times
    du = np.zeros([len(unfused_in[0])])
    vu = np.zeros([len(unfused_in[0])])
    eu = np.zeros([len(unfused_in[0])])

    for i in range(len(unfused_in[0])):
        vu[i] = nvec[unfused_in[0][i]]
        eu[i] = np.log2(nelem[unfused_in[1][i]])
        du[i] = 1

    if (args.debug):
        print(vu)
        print(eu)

    # --------------------------------------------------------------------------
    # Output ratios
    # --------------------------------------------------------------------------

    # output the matrix of ratios in reverse row order to match the heatmap
    print(args.op)

    # print(avg_fused)
    # for i in reversed(range(len(nvec))):
    #     print('%2d' % int(i+1), str(avg_fused[i]).replace('\n', ''))
    # print

    # print(avg_unfused)
    # for i in reversed(range(len(nvec))):
    #     print('%2d' % int(i+1), str(avg_unfused[i]).replace('\n', ''))
    # print

    # print(NVEC)
    # print(NELM)
    # print(avg_ratio)
    for i in reversed(range(len(nvec))):
        print('%2d' % int(i+1), str(avg_ratio[i]).replace('\n', ''))
    print

    # --------------------------------------------------------------------------
    # Heat Map
    # --------------------------------------------------------------------------
    if (args.heatmap):

        x = np.arange(len(nelem)+1)-0.5 # x = log2(number of elements) = 0,1,2,...
        y = np.arange(len(nvec)+1)+1.5  # y = number of vectors = 2,3,4,...
        # y = np.arange(len(nvec)+1)+0.5  # y = number of vectors = 1,2,3,...
        X, Y = np.meshgrid(x, y)

        if (args.debug):
            print(x)
            print(y)

        # center the color bar around 1 (if possible)
        rmax = np.amax(avg_ratio)
        rmin = np.amin(avg_ratio)

        ext = 'neither'
        if (rmin > 1):
            cmap='Reds'
            norm = mpl.colors.Normalize(vmin=rmin, vmax=min(rmax,2))
            v = np.linspace(rmin, min(rmax,2), 10, endpoint=True)
            if (rmax > 2):
                ext = 'max'
        else:
            cmap='seismic'
            if (rmax-1 > 1):
                rrange = 1
                ext = 'max'
            else:
                rrange = max(abs(rmax-1),abs(rmin-1))

            v1 = np.linspace(1-rrange, 1, 5, endpoint=True)
            v2 = np.linspace(1, 1+rrange, 5, endpoint=True)
            v = np.append(v1,v2[1:])
            norm = mpl.colors.Normalize(vmin=1-rrange, vmax=1+rrange)

        # plot heatmap
        plt.pcolormesh(X, Y, avg_ratio, cmap=cmap, norm=norm)
        clb = plt.colorbar(ticks=v, extend=ext)
        clb.ax.set_title('Max = {0:.2f}\nMin = {1:.2f}'.format(rmax,rmin))

        # aff markers to indicate if the average time falls in a confidence interval
        plt.scatter(ef,vf,s=40,marker='^',c=df,label='fused')
        plt.scatter(eu,vu,s=40,marker='v',c=du,label='unfused')
        plt.legend(loc=9, bbox_to_anchor=(0.5, -0.1), ncol=2)

        # add legend for scatter plot
        art = []
        lgd = plt.legend(loc='lower right', bbox_to_anchor=(1.34, -0.17))
        art.append(lgd)

        # add labels and title
        plt.xticks(np.log2(nelem))
        plt.yticks(nvec)
        plt.xlabel('log2(num elements)')
        plt.ylabel('num vectors')
        plt.title('avg fused time / avg unfused time \n'+args.op)

        # display or save figure
        if (args.show):
            plt.show()
        else:
            plt.savefig(args.op+'-heatmap.pdf',
                        additional_artists=art,
                        bbox_inches="tight")
        plt.close()

    # --------------------------------------------------------------------------
    # Time vs Number of Elements Plots
    # --------------------------------------------------------------------------
    if (args.timevelem):

        colors = ['#000000','#a6cee3','#1f78b4','#b2df8a','#33a02c',
                  '#fb9a99','#e31a1c','#fdbf6f','#ff7f00','#cab2d6',
                  '#6a3d9a','#ffff99','#b15928']

        hatch = [ '/','\\','-','+','x','o','O','.','*']

        # --------------------------------------------------------------------------
        # Combined Number of Vectors Plots
        # --------------------------------------------------------------------------
        fig = plt.figure()
        ax = fig.add_subplot(111)

        for nv in nvec:

            i = nvec.index(nv)

            if (args.loglog):
                ax.loglog(nelem, avg_fused[i],
                          color=colors[i], linestyle='-', label=nv)
                ax.loglog(nelem, avg_unfused[i],
                          color=colors[i], linestyle='--', label=None)
            else:
                ax.plot(nelem, avg_fused[i],
                        color=colors[i], linestyle='-', label=nv)
                ax.plot(nelem, avg_unfused[i],
                        color=colors[i], linestyle='--', label=None)

            # plot confidence interval
            ax.fill_between(nelem, lower_fused[i], upper_fused[i],
                            color=colors[i], alpha=0.3)
            ax.fill_between(nelem, lower_unfused[i], upper_unfused[i],
                            color=colors[i], hatch='.', alpha=0.3)

        ax.legend()
        ax.grid()

        plt.title('Average Time Fused vs Unfused \n'+args.op)
        plt.xlabel('vector length')
        plt.ylabel('time (s)')

        if (args.show):
            plt.show()
        else:
            if (args.loglog):
                fname=args.op+'-nvec-all-loglog.pdf'
            else:
                fname=args.op+'-nvec-all.pdf'
                plt.ticklabel_format(axis='both',style='sci')
            plt.savefig(fname)
        plt.close()

        # --------------------------------------------------------------------------
        # Individual Number of Vectors Plots
        # --------------------------------------------------------------------------

        for nv in nvec:
            fig = plt.figure()
            ax  = fig.add_subplot(111)
            idx = nvec.index(nv)

            # plot run times
            if (args.loglog):
                ax.loglog(nelem, avg_fused[idx],
                          color='red', linestyle='-', label='Fused')
                ax.loglog(nelem, avg_unfused[idx],
                          color='blue', linestyle='--', label='Unfused')
            else:
                ax.plot(nelem, avg_fused[idx],
                        color='red', linestyle='-', label='Fused')
                ax.plot(nelem, avg_unfused[idx],
                        color='blue', linestyle='--', label='Unfused')

            # plot confidence intervals
            ax.fill_between(nelem, lower_fused[idx], upper_fused[idx],
                            color='red', alpha=0.2)
            ax.fill_between(nelem, lower_unfused[idx], upper_unfused[idx],
                            color='blue', hatch='.', alpha=0.2)

            ax.legend()
            ax.grid()

            plt.title('Average Time Fused vs Unfused with '+str(nv)+' vectors\n'+args.op)
            plt.xlabel('vector length')
            ax.set_ylabel('time (s)')

            if (args.show):
                plt.show()
            else:
                if (args.loglog):
                    fname=args.op+'-nvec-'+str(nv)+'-loglog.pdf'
                else:
                    fname=args.op+'-nvec-'+str(nv)+'.pdf'
                    plt.ticklabel_format(axis='both',style='sci')
                plt.savefig(fname)
            plt.close()

# ===============================================================================

if __name__ == "__main__":
    main()

# EOF




#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Script to update example output files for failed tests
#
# Example usage:
#   $ ./updateOutFiles.py sundials/build sundials/examples
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# main routine
# -----------------------------------------------------------------------------
def main():

    import argparse
    import sys, os, shutil

    parser = argparse.ArgumentParser(description='Update output files')

    parser.add_argument('build', type=str,
                        help='Full path to build directory to read from')
    parser.add_argument('examples', type=str,
                        help='Full path to examples directory to write to')
    parser.add_argument('--all','-a', action='store_true',
                        help='Update all output files')
    parser.add_argument('--verbose','-v', action='store_true',
                        help='Enable verbose output')


    # parse command line args
    args = parser.parse_args()

    # check inputs
    if (not os.path.isdir(args.examples)):
        print("Error: could not find {}".format(args.examples))
        return -1

    if (not os.path.isdir(args.build)):
        print("Error: could not find {}".format(args.build))
        return -1

    # check that the output directory exists
    output = os.path.join(args.build, "Testing", "output")
    if (not os.path.isdir(output)):
        print("Error: could not find {}".format(output))
        return -1

    # create a list of all test run or just the failed tests
    tests = []

    if (args.all):
        # get names of all .out files
        for f in os.listdir(output):
            if f.endswith(".out"):
                tests.append(f)
    else:
        failed = os.path.join(args.build, "Testing", "Temporary", "LastTestsFailed.log")

        # extract test names from list and append .out
        with open(failed, 'r') as f:
            for line in f:
                tests.append(line.split(':')[1].rstrip() + ".out")

    # if failed tests were found update the output files
    if tests:
        for t in tests:
            found = False
            for root, dirs, files in os.walk(args.examples):
                if t in files:
                    if (args.verbose):
                        print("Updating: {}".format(t))
                    shutil.copy(os.path.join(output, t), os.path.join(root, t))
                    found = True
                    break
            if not found:
                print("Warning: did not find {}".format(t))

# -----------------------------------------------------------------------------
# run the main routine
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    import sys
    sys.exit(main())



#!/usr/bin/env python
# ----------------------------------------------------------------
# Programmer(s): Daniel R. Reynolds @ SMU
# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
# matplotlib-based plotting script
# ----------------------------------------------------------------

# imports
import glob
import sys
import pylab as plt
import numpy as np

# load mesh data file
mesh = np.loadtxt('mesh.txt', dtype=np.double)

# load output time file
times = np.loadtxt('t.000000.txt', dtype=np.double)

# load solution data files
ufiles = glob.glob('u.' + ('[0-9]'*6) + '.txt'); ufiles.sort()
vfiles = glob.glob('v.' + ('[0-9]'*6) + '.txt'); vfiles.sort()
wfiles = glob.glob('w.' + ('[0-9]'*6) + '.txt'); wfiles.sort()
udata = np.loadtxt(ufiles[0], dtype=np.double)
vdata = np.loadtxt(vfiles[0], dtype=np.double)
wdata = np.loadtxt(wfiles[0], dtype=np.double)
for idx in range(1,len(ufiles)):
    udata = np.hstack((udata, np.loadtxt(ufiles[idx], dtype=np.double)))
    vdata = np.hstack((vdata, np.loadtxt(vfiles[idx], dtype=np.double)))
    wdata = np.hstack((wdata, np.loadtxt(wfiles[idx], dtype=np.double)))

# determine number of time steps, mesh size
nt,nx = np.shape(udata)

# determine min/max values
umin = 0.9*udata.min()
umax = 1.1*udata.max()
vmin = 0.9*vdata.min()
vmax = 1.1*vdata.max()
wmin = 0.9*wdata.min()
wmax = 1.1*wdata.max()
xmax = mesh.max()
minval = np.array([umin, vmin, wmin]).min()
maxval = np.array([umax, vmax, wmax]).max()

# generate plots of results
for tstep in range(nt):

    # set string constants for output plots, current time, mesh size
    pname = 'solution.' + repr(tstep).zfill(3) + '.png'
    tstr  = repr(tstep)
    nxstr = repr(nx)

    # plot current solution and save to disk
    plt.figure(1)
    plt.plot(mesh,udata[tstep,:],label="u")
    plt.plot(mesh,vdata[tstep,:],label="v")
    plt.plot(mesh,wdata[tstep,:],label="w")
    plt.xlabel('x')
    plt.ylabel('solution')
    plt.title('Solutions at output ' + tstr + ', mesh = ' + nxstr)
    plt.axis((0.0, xmax, minval, maxval))
    plt.grid()
    plt.legend(loc='upper right', shadow=True)
    plt.savefig(pname)
    plt.close()

# set string constants for output plots, current time, mesh size
pname = 'solution_at_x0.png'
xstr = repr(mesh[0])

# plot current solution and save to disk
plt.figure(1)
plt.plot(times,udata[:,0],label="u")
plt.plot(times,vdata[:,0],label="v")
plt.plot(times,wdata[:,0],label="w")
plt.xlabel('t')
plt.ylabel('solution')
plt.title('Solutions at output at x = '+xstr)
plt.axis((times[0], times[-1], minval, maxval))
plt.grid()
plt.legend(loc='upper right', shadow=True)
plt.savefig(pname)
plt.close()


##### end of script #####



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# matplotlib-based plotting script for cvsPendulum_dns.c example
# -----------------------------------------------------------------------------

# imports
import argparse
import numpy as np
import matplotlib.pyplot as plt

# command line options
parser = argparse.ArgumentParser(description='Plots cvsPendulum_dns output')
parser.add_argument('sfile', type=str,
                    help='solution output file to read')

# parse inputs
args = parser.parse_args()

# read solution output file
data = np.loadtxt(args.sfile, dtype=np.double)

# extract times, positions, and velocities
t  = data[:, 0]
x  = data[:, 1]
y  = data[:, 2]
vx = data[:, 3]
vy = data[:, 4]

# read reference solution output file
ref = np.loadtxt("cvsPendulum_dns_ref.txt", dtype=np.double)

# extract positions and velocities
xr  = ref[:, 1]
yr  = ref[:, 2]
vxr = ref[:, 3]
vyr = ref[:, 4]

# lower half of unit circle
tt = np.linspace(np.pi, 2*np.pi, 10000)
xt = np.cos(tt)
yt = np.sin(tt)

# plot solution in xy plane
fig, ax = plt.subplots()
ax.axhline(y=0, color='black', linestyle='--', label=None)
ax.axvline(x=0, color='black', linestyle='--', label=None)
plt.plot(xt, yt, color='black', linestyle='--', label=None)
plt.scatter(x, y, color='red', label='comp')
plt.scatter(xr, yr, color='blue', label='ref')

plt.xlabel('x')
plt.ylabel('y')
plt.title('Pendulum')
ax.set_aspect('equal')
plt.legend(loc='lower right')

# plot position over time
fig, ax = plt.subplots()
ax.axhline(y=0, color='black', linestyle='--')
plt.plot(t, x, label='x')
plt.plot(t, y, label='y')

plt.xlabel('t')
plt.ylabel('position')
plt.title('Pendulum Position')
plt.legend()

# plot velocity over time
fig, ax = plt.subplots()
ax.axhline(y=0, color='black', linestyle='--')
plt.plot(t, vx, label='$v_x$')
plt.plot(t, vy, label='$v_y$')

plt.xlabel('t')
plt.ylabel('velocity')
plt.title('Pendulum Velocity')
plt.legend()

# display plots
plt.show()

##### end of script #####



#!/usr/bin/env python
# ----------------------------------------------------------------
# Programmer(s): Daniel R. Reynolds @ SMU
# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
# matplotlib-based plotting script for heat1D.c example

# imports
import sys
import pylab as plt
import numpy as np

# load mesh data file
mesh = np.loadtxt('heat_mesh.txt', dtype=np.double)

# load solution data file
data = np.loadtxt('heat1D.txt', dtype=np.double)

# determine number of time steps, mesh size
nt,nx = np.shape(data)

# determine maximum temperature
maxtemp = 1.1*data.max()

# generate plots of results
for tstep in range(nt):

    # set string constants for output plots, current time, mesh size
    pname = 'heat1d.' + repr(tstep).zfill(3) + '.png'
    tstr  = repr(tstep)
    nxstr = repr(nx)

    # plot current solution and save to disk
    plt.figure(1)
    plt.plot(mesh,data[tstep,:])
    plt.xlabel('x')
    plt.ylabel('solution')
    plt.title('u(x) at output ' + tstr + ', mesh = ' + nxstr)
    plt.axis((0.0, 1.0, 0.0, maxtemp))
    plt.grid()
    plt.savefig(pname)
    plt.close()


##### end of script #####



# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

import sys, os
sys.path.append(os.path.dirname(os.path.abspath('../../../shared/versions.py')))
from versions import *

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '4.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.todo', 'sphinx.ext.mathjax']


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Example Programs for ARKODE'
copyright = u'2012-2021, Daniel R. Reynolds, David J. Gardner, Carol S. Woodward, and Cody J. Balos, release number LLNL-SM-668082'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '{arkode_version}'.format(arkode_version=arkode_version)
sun_version = '{sundials_version}'.format(sundials_version=sundials_version)

# Set the date format (full-month-name day, full-year)
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'
highlight_language = "c"

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

rst_epilog = """
.. |YEAR| replace:: {year}
.. |CVODE_VERSION| replace:: {cvode_version}
.. |CVODES_VERSION| replace:: {cvodes_version}
.. |ARKODE_VERSION| replace:: {arkode_version}
.. |IDA_VERSION| replace:: {ida_version}
.. |IDAS_VERSION| replace:: {idas_version}
.. |KINSOL_VERSION| replace:: {kinsol_version}
""".format(year = year,
cvode_version = cvode_version,
cvodes_version = cvodes_version,
arkode_version = arkode_version,
ida_version = ida_version,
idas_version = idas_version,
kinsol_version = kinsol_version
)


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

##########
# agogo theme and configuration
# html_theme = 'agogo'
# html_theme_options = {
#  "bodyfont": "Arial",
#  "headerfont": "Arial",
#  "pagewidth": "70em",
#  "documentwidth": "50em",
#  "sidebarwidth": "20em",
#  "bgcolor": "#ffffff",
#  "headerbg": "#ffffff",
#  "footerbg": "#ffffff",
#  "linkcolor": "",
#  "headercolor1": "",
#  "headercolor2": "",
#  "headerlinkcolor": "",
#  "textalign": "justify"
#}
#html_theme_path = ["."]


##########
# custom bootstrap theme and configuration

sys.path.append(os.path.abspath('_themes'))
html_theme = 'bootstrap'
html_theme_path = ['_themes']
html_theme_options = {
    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 1,

    # HTML navbar class (Default: "navbar") to attach to <div>.
    # For black navbar, do "navbar navbar-inverse"
    'navbar_class': "navbar navbar-inverse",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "true",

    # Location of link to source.
    # Options are "nav" (default), "footer".
    'source_link_position': "nav",
}


##########
# bootstrap theme and configuration

# import sphinx_bootstrap_theme
# html_theme = 'bootstrap'
# html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()
# html_theme_options = {
#     # Global TOC depth for "site" navbar tab. (Default: 1)
#     # Switching to -1 shows all levels.
#     'globaltoc_depth': 1,

#     # HTML navbar class (Default: "navbar") to attach to <div>.
#     # For black navbar, do "navbar navbar-inverse"
#     'navbar_class': "navbar navbar-inverse",

#     # Fix navigation bar to top of page?
#     # Values: "true" (default) or "false"
#     'navbar_fixed_top': "true",

#     # Location of link to source.
#     # Options are "nav" (default), "footer".
#     'source_link_position': "nav",
# }


#############

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'ARKODEExampleDoc'


# -- Options for LaTeX output --------------------------------------------------

# 1. the rst file name used to generate the LaTeX file
# 2. the name of the LaTeX file to generate (and the resulting PDF file name)
# 3. the document title
# 4. text for \author
# 5. the LaTeX theme
# 6. include the file from 1. in the output
tex_author = r'''
    Daniel R. Reynolds\\
    {\em Department of Mathematics} \\
    {\em Southern Methodist University}
    '''

latex_documents = [('index', 'ark_examples.tex', project,
                    tex_author, 'manual', False)]

# make sure the doc logo gets copied to the build directory
latex_logo = 'figs/doc_logo_blue.pdf'

# LaTeX customizations
latex_elements = {
# paper size option of the document class
'papersize': 'letterpaper',
# font size option of the document class
'pointsize': '10pt',
# set the version number/release name
'releasename' : version,
# arguments to the sphinxsetup macro
'sphinxsetup':
    # the color for titles
    'TitleColor={RGB}{0,0,0},'+
    # disable frames around code-blocks
    'verbatimwithframe=false,'+
    # do not wrap long lines in code-blocks
    'verbatimwrapslines=false,'+
    # background color for code-blocks
    'VerbatimColor={RGB}{240.0,240.0,240.0},'+
    # font used by heading
    'HeaderFamily=\\rmfamily\\bfseries',
# disable the fncychap package
'fncychap':'',
# figure alignment options
'figure_align': 'htbp',
# additional preamble content
'preamble': r'''
% =====================================================
% Start custom preamble (see latex_elements in conf.py)
% =====================================================

% Use ragged-right for the whole document
\usepackage[document]{ragged2e}

% Specify depths for section numbering and table of contents
\setcounter{tocdepth}{1}
\setcounter{secnumdepth}{3}

% Link a footnote to its location in the text
\usepackage{footnotebackref}

% Add new command for SUNDIALS version
'''
+
r'\newcommand{\sunreleasename}{' + sun_version + r'}'
+
r'''

% ===================================================
% End custom preamble (see latex_elements in conf.py)
% ===================================================
''',
# custom maketitle
'maketitle': r'''
% =======================================================
% Start custom cover page (see latex_elements in conf.py)
% =======================================================

\makeatletter

% Start roman numbering
\pagenumbering{Roman}

% Title page
\begin{titlepage}
  \newpage
  \null
  \vskip 2em%
  \begin{center}%
    \let \footnote \thanks
    {\huge \rmfamily \@title \space \releasename \par}
    {\Large \rmfamily SUNDIALS \space \sunreleasename \par}
    \vskip 3.0em%
    {\large \lineskip .5em%
     \begin{tabular}[t]{c}%
       \@author
     \end{tabular}\par}%
    \vskip 1em%
    {\large \@date \par}%
    \vfill
    {\includegraphics[width=0.5\textwidth]{doc_logo_blue}}
    \vfill
    %{\large \rmfamily LLNL-SM-668082}
    \vfill
  \end{center}
  \par
  \vskip 1.5em
\end{titlepage}

\makeatother

\clearpage

% Disclaimer
\thispagestyle{empty}% no number of this page
\vglue5\baselineskip
\begin{center}
  {\bf DISCLAIMER}
\end{center}
\noindent
This document was prepared as an account of work sponsored by an agency of
the United States government. Neither the United States government nor
Lawrence Livermore National Security, LLC, nor any of their employees makes
any warranty, expressed or implied, or assumes any legal liability or responsibility
for the accuracy, completeness, or usefulness of any information, apparatus, product,
or process disclosed, or represents that its use would not infringe privately owned rights.
Reference herein to any specific commercial product, process, or service by trade name,
trademark, manufacturer, or otherwise does not necessarily constitute or imply its endorsement,
recommendation, or favoring by the United States government or Lawrence Livermore National
Security, LLC. The views and opinions of authors expressed herein do not necessarily state
or reflect those of the United States government or Lawrence Livermore National Security, LLC,
and shall not be used for advertising or product endorsement purposes.

\vskip2\baselineskip
\noindent
This work was performed under the auspices of the U.S. Department of Energy by Lawrence
Livermore National Laboratory under Contract DE-AC52-07NA27344.
\vfill
\begin{center}
  Approved for public release; further dissemination unlimited
\end{center}

\clearpage

% Contributors
\thispagestyle{empty}% no number of this page
\vglue5\baselineskip
\begin{center}
{\bf CONTRIBUTORS}
\end{center}
\noindent
The SUNDIALS library has been developed over many years by a number of
contributors. The current SUNDIALS team consists of Cody J. Balos,
David J. Gardner, Alan C. Hindmarsh, Daniel R. Reynolds, and
Carol S. Woodward. We thank Radu Serban for significant and critical past
contributions.\\
\vskip 2em%
\noindent
Other contributors to SUNDIALS include: James Almgren-Bell, Lawrence E. Banks,
Peter N. Brown, George Byrne, Rujeko Chinomona, Scott D. Cohen, Aaron Collier,
Keith E. Grant, Steven L. Lee, Shelby L. Lockhart, John Loffeld, Daniel McGreer,
Slaven Peles, Cosmin Petra, H. Hunter Schwartz, Jean M. Sexton,
Dan Shumaker, Steve G. Smith, Allan G. Taylor, Hilari C. Tiedeman, Chris White,
Ting Yan, and Ulrike M. Yang.
\clearpage

% clear empty double page
\newpage{\pagestyle{empty}\cleardoublepage}

% Start arabic numbering
\pagenumbering{arabic}

% =====================================================
% End custom cover page (see latex_elements in conf.py)
% =====================================================
'''
}


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'arkode_example', u'ARKODE Example Documentation',
     [u'Daniel R. Reynolds'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'ARKODE_example', u'ARKODE Example Documentation',
   u'Daniel R. Reynolds', 'ARKODE_example',
   'Example programs for the ARKODE time integration package for multi-rate systems of ordinary differntial equations.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'



#!/usr/bin/env python
# ------------------------------------------------------------------------------
# Programmer(s):  Daniel R. Reynolds @ SMU
#                 David J. Gardner @ LLNL
# ------------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ------------------------------------------------------------------------------
# matplotlib-based plotting script for the parallel ark_heat2D_p examples
# ------------------------------------------------------------------------------

# imports
import sys, os
import shlex
import numpy as np
from pylab import *
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.pyplot as plt

# ------------------------------------------------------------------------------

# read MPI root process problem info file
infofile = 'heat2d_info.00000.txt'

with open(infofile) as fn:

    # read the file line by line
    for line in fn:

        # split line into list
        text = shlex.split(line)

        # x-direction upper domian bound
        if "xu" in line:
            xu = float(text[1])
            continue

        # y-direction upper domain bound
        if "yu" in line:
            yu = float(text[1])
            continue

        # nodes in the x-direction (global)
        if "nx" in line:
            nx = int(text[1])
            continue

        # nodes in the y-direction (global)
        if "ny" in line:
            ny = int(text[1])
            continue

        # total number of MPI processes
        if "np"in line:
            nprocs = int(text[1])
            continue

        # number of output times
        if "nt" in line:
            nt = int(text[1])
            continue

# ------------------------------------------------------------------------------

# load subdomain information, store in table
subdomains = np.zeros((nprocs,4), dtype=np.int)

for i in range(nprocs):

    infofile = 'heat2d_info.' + repr(i).zfill(5) + '.txt'

    with open(infofile) as fn:

        # read the file line by line
        for line in fn:

            # split line into list
            text = shlex.split(line)

            # x-direction starting index
            if "is" in line:
                subdomains[i,0] = float(text[1])
                continue

            # x-direction ending index
            if "ie" in line:
                subdomains[i,1] = float(text[1])
                continue

            # y-direction starting index
            if "js" in line:
                subdomains[i,2] = float(text[1])
                continue

            # y-direction ending index
            if "je" in line:
                subdomains[i,3] = float(text[1])
                continue

# ------------------------------------------------------------------------------

# check if the error was output
fname = 'heat2d_error.00000.txt'

if os.path.isfile(fname):
    plottype = ['solution', 'error']
else:
    plottype = ['solution']

for pt in plottype:

    # fill array with data
    time   = np.zeros(nt)
    result = np.zeros((nt, ny, nx))

    for i in range(nprocs):

        datafile = 'heat2d_' + pt + '.' + repr(i).zfill(5) + '.txt'

        # load data
        data = np.loadtxt(datafile, dtype=np.double)

        if (np.shape(data)[0] != nt):
            sys.exit('error: subdomain ' + i + ' has an incorrect number of time steps')

        # subdomain indices
        istart = subdomains[i,0]
        iend   = subdomains[i,1]
        jstart = subdomains[i,2]
        jend   = subdomains[i,3]
        nxl    = iend - istart + 1
        nyl    = jend - jstart + 1

        # extract data
        for i in range(nt):
            time[i] = data[i,0]
            result[i,jstart:jend+1,istart:iend+1] = np.reshape(data[i,1:], (nyl,nxl))

    # determine extents of plots
    maxtemp = 1.1 * result.max()
    mintemp = 0.9 * result.min()

    # set x and y meshgrid objects
    xspan = np.linspace(0.0, xu, nx)
    yspan = np.linspace(0.0, yu, ny)
    X,Y   = np.meshgrid(xspan, yspan)

    nxstr = repr(nx)
    nystr = repr(ny)

    # generate plots
    for tstep in range(nt):

        # set string constants for output plots, current time, mesh size
        pname = 'heat2d_surf_' + pt + '.' + repr(tstep).zfill(3) + '.png'
        tstr  = str(time[tstep])

        # plot surface and save to disk
        fig = plt.figure(1)
        ax = fig.add_subplot(111, projection='3d')

        ax.plot_surface(X, Y, result[tstep,:,:], rstride=1, cstride=1,
                        cmap=cm.jet, linewidth=0, antialiased=True, shade=True)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlim((mintemp, maxtemp))
        ax.view_init(20,45)
        if (pt == 'solution'):
            title('u(x,y) at t = ' + tstr)
        else:
            title('error(x,y) at t = ' + tstr)
        savefig(pname)
        plt.close()

##### end of script #####



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Script to plot data from a 2-dimensional problem on a rectangular domain with
# uniform grid spacing.
#
# The input data file(s) must contain an M x N matrix of values where M is the
# number of output times and the N = 1 + V * X * Y columns contain the output
# data where V is the number of solution variables, X is the number of nodes in
# the x-direction, and Y is the number of nodes in the y-direction.
#
# The first column of each row is the output time for the solution data in the
# remaining columns. The output data must be ordered so the solution variables
# at a given location grouped together and the spatial node ordering varies
# first over x and then y. For example if V = 2, X = 2, and Y = 2 then the
# output file would contain the data
#
# t_0      u_0,0   v_0,0   u_1,0   v_1,0   u_0,1   v_0,1   u_1,1   v_1,1
# t_1      u_0,0   v_0,0   u_1,0   v_1,0   u_0,1   v_0,1   u_1,1   v_1,1
#  .         .       .       .       .       .       .       .       .
#  .         .       .       .       .       .       .       .       .
#  .         .       .       .       .       .       .       .       .
# t_{M-1}  u_0,0   v_0,0   u_1,0   v_1,0   u_0,1   v_0,1   u_1,1   v_1,1
#
# where u_i,j and v_i,j are the solution components at node (i,j). In general,
# the n-th solution component at node (i,j) is located at column index
# 1 + V * (X * j + i) + n with n = 0,...,V-1, i = 0,...,X-1, and j = 0,...,Y-1.
#
# Additionally data file(s) should contain the following header comment block
# describing the output data (note comment lines begin with #):
#
#   # nprocs  <number of mpi processes>
#   # nvar    <number of solution variables>
#   # nt      <number of output times>
#   # nx      <number of nodes in the x-direction>
#   # xl      <x-direction lower bound>
#   # xu      <x-direction upper bound>
#   # is      <subdomain global starting x-index>
#   # ie      <subdomain global ending x-index>
#   # ny      <number of nodes in the y-direction>
#   # yl      <y-direction lower bound>
#   # yu      <y-direction upper bound>
#   # js      <subdomain global starting y-index>
#   # je      <subdomain global ending y-index>
#
# With the exception of the subdomain indices, if any of the above values are
# not provided in the header the script will attempt to deduce the necessary
# values and will print a warning or halt with an error.
#
# The comment block may optionally contain additional entries used in creating
# plots:
#   # title    <plot title>
#   # varnames <variables names>
#
# Example usage:
#   plot_data_2d.py data_file.out
#   plot_data_2d.py data_file_proc_0.out data_file_proc_1.out
#   plot_data_2d.py data_file.out --plottype surface-ani
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# main routine
# -----------------------------------------------------------------------------
def main():

    import sys
    import argparse

    parser = argparse.ArgumentParser(description='''Plot 2D data files''')

    # List of input data files

    parser.add_argument('datafiles', type=str, nargs='+',
                        help='Data files to plot')

    # Plot type options

    group = parser.add_argument_group('Plot Options',
                                      '''Options to specify the type of plot to
                                      generate and what data to plot''')

    group.add_argument('--plottype', type=str,
                       choices=['surface', 'surface-ani',
                                'contour', 'contour-ani',
                                'slice', 'point'],
                       default='surface',
                       help='''Set the plot type''')

    group.add_argument('--plotvars', type=int, nargs='+',
                       help='''Variable indices to plot''')

    group.add_argument('--plottimes', type=int, nargs='+',
                       help='''Time indices to plot''')

    # Slice plot options

    group = parser.add_argument_group('Slice Plot Options',
                                      '''Options specific to the slice plot
                                      type''')

    group.add_argument('--slicetype', type=str, default='var',
                       choices=['var', 'time'],
                       help='''The slice plot type''')

    mxgroup = group.add_mutually_exclusive_group()

    mxgroup.add_argument('--yslice', type=int, default=-1,
                         help='''y index to plot''')

    mxgroup.add_argument('--xslice', type=int, default=-1,
                         help='''x index to plot''')

    # Point plot options

    group = parser.add_argument_group('Point Plot Options',
                                      '''Options specific to the point plot
                                      type''')

    group.add_argument('--point', type=int, nargs=2, default=[0, 0],
                       help='''x and y index to plot''')

    # Output options

    group = parser.add_argument_group('Output Options',
                                      '''Options for saving plots''')

    group.add_argument('--save', action='store_true',
                       help='''Save figure to file''')

    group.add_argument('--prefix', type=str,
                       help='''File name prefix for saving the figure''')

    group.add_argument('--merge', action='store_true',
                       help='''Merge PDF output files into a single file''')

    # Figure options

    group = parser.add_argument_group('Figure Options',
                                      '''Options to specify various figure
                                      properties''')

    group.add_argument('--labels', type=str, nargs='+',
                       help='''Data labels for the plot legend''')

    group.add_argument('--title', type=str,
                       help='''Plot title''')

    group.add_argument('--xlabel', type=str,
                       help='''x-axis label''')

    group.add_argument('--ylabel', type=str,
                       help='''y-axis label''')

    group.add_argument('--zlabel', type=str,
                       help='''z-axis label''')

    group.add_argument('--grid', action='store_true',
                       help='''Add grid to plot''')

    # Debugging options

    parser.add_argument('--debug', action='store_true',
                        help='Enable debugging')

    # parse command line args
    args = parser.parse_args()

    # create dictionary with header info
    info = read_header(args)

    # create matrix of subdomain info
    subdomains = read_subdomains(args, info)

    # read data
    time, xvals, yvals, zdata = read_data(args, info, subdomains)

    # setup plot info
    plot_settings(args, info, time, xvals, yvals, zdata)

    # Create plots
    if args.plottype == 'surface':
        plot_surface(args, info, time, xvals, yvals, zdata)

    if args.plottype == 'surface-ani':
        plot_surface_ani(args, info, time, xvals, yvals, zdata)

    if args.plottype == 'contour':
        plot_contour(args, info, time, xvals, yvals, zdata)

    if args.plottype == 'contour-ani':
        plot_contour_ani(args, info, time, xvals, yvals, zdata)

    if args.plottype == 'slice':

        # slice data
        if (args.yslice > -1) and (args.yslice < info['ny']):
            svals = xvals
            sdata = zdata[:, args.yslice, :, :]
            if args.xlabel:
                hlabel = args.xlabel
            else:
                hlabel = 'x'
            suffix = " at y = {:.4f}".format(yvals[args.yslice])
        elif (args.xslice > -1) and (args.xslice < info['nx']):
            svals = yvals
            sdata = zdata[:, :, args.xslice, :]
            if args.ylabel:
                hlabel = args.ylabel
            else:
                hlabel = 'y'
            suffix = " at x = {:.4f}".format(xvals[args.xslice])
        else:
            print("ERROR: invalid xslice or yslice option")
            sys.exit()

        if args.slicetype == 'var':
            plot_slice_vars(args, info, time, svals, sdata, hlabel, suffix)
        else:
            plot_slice_time(args, info, time, svals, sdata, hlabel, suffix)

    if args.plottype == 'point':

        # point data
        pdata = zdata[:, args.point[1], args.point[0], :]
        suffix = " at x = {:.4f}, y = {:.4f}".format(xvals[args.point[0]],
                                                     yvals[args.point[1]])

        plot_point(args, info, time, pdata, suffix)


# -----------------------------------------------------------------------------
# function to print info dictionary
# -----------------------------------------------------------------------------


def print_info(info):

    print("Info dictionary:")
    for k, v in info.items():
        print("  ", k, v)


# -----------------------------------------------------------------------------
# function to read data file header
# -----------------------------------------------------------------------------


def read_header(args):

    import sys
    import shlex
    import numpy as np

    # initialize dictionary of header info variables to None
    keys = ['title', 'varnames', 'nprocs', 'nvar', 'nt', 'nx', 'xl', 'xu',
            'ny', 'yl', 'yu']

    info = dict()
    for k in keys:
        info[k] = None

    # read the first input file and extract info from the header
    with open(args.datafiles[0]) as fn:

        # read the file line by line
        for line in fn:

            # skip empty lines
            if not line.strip():
                continue

            # exit after reading initial comment lines
            if "#" not in line:
                break

            # split line into list
            text = shlex.split(line)

            # plot title
            if "title" in line:
                info['title'] = " ".join(text[2:])
                continue

            # plot variable names
            if "vars" in line:
                info['varnames'] = text[2:]
                continue

            # total number of processes
            if "nprocs" in line:
                info['nprocs'] = int(text[2])
                continue

            # number of variables (at each spatial node)
            if "nvar" in line:
                info['nvar'] = int(text[2])
                continue

            # number of output times
            if "nt" in line:
                info['nt'] = int(text[2])
                continue

            # the global number of nodes in the x-direction, the x lower bound
            # (west) and the x upper bound (east)
            if "nx" in line:
                info['nx'] = int(text[2])
                continue
            if "xl" in line:
                info['xl'] = float(text[2])
                continue
            if "xu" in line:
                info['xu'] = float(text[2])
                continue

            # the global number of nodes in the y-direction, the y lower bound
            # (south) and the y upper bound (north)
            if "ny" in line:
                info['ny'] = int(text[2])
                continue
            if "yl" in line:
                info['yl'] = float(text[2])
                continue
            if "yu" in line:
                info['yu'] = float(text[2])
                continue

    # load data to deduce values and perform sanity checks
    data = np.loadtxt(args.datafiles[0], dtype=np.double)

    # try to fill in missing values
    if info['nvar'] is None:
        info['nvar'] = 1
        print("WARNING: nvar not provided. Using nvar = 1")

    if info['nt'] is None or info['nx'] is None or info['ny'] is None:

        # check if data exists
        if data.ndim != 2:
            print("ERROR: data file is not 2d")
            sys.exit()

        # number of output times
        if info['nt'] is None:
            info['nt'] = np.shape(data)[0]

        # number of spatial nodes
        if info['nx'] is None or info['ny'] is None:
            col = np.shape(data)[1] - 1  # exclude output times
            if info['nx'] is None and info['ny'] is not None:
                info['nx'] = col // (info['nvar'] * info['ny'])
            elif info['nx'] is not None and info['ny'] is None:
                info['ny'] = col // (info['nvar'] * info['nx'])
            else:
                info['nx'] = int(np.sqrt(col // info['nvar']))
                info['ny'] = info['nx']
                print("WARNING: nx and ny not provided. Using nx = ny =",
                      info['nx'])

    # sanity checks
    if info['nt'] != np.shape(data)[0]:
        print("ERROR: nt != nrows", info['nt'], np.shape(data)[0])
        sys.exit()

    if (info['nvar'] * info['nx'] * info['ny']) != (np.shape(data)[1] - 1):
        print("ERROR: nvar * nx * ny != ncols - 1")
        sys.exit()

    # check x-dimension lower and upper bounds
    if info['xl'] is None:
        print("WARNING: xl not provided, using xl = 0")
        info['xl'] = 0.0

    if info['xu'] is None:
        print("WARNING: xu not provided, using xu = 1")
        info['xu'] = 1.0

    # check y-dimension lower and upper bounds
    if info['yl'] is None:
        print("WARNING: yl not provided, using yl = 0")
        info['yl'] = 0.0

    if info['yu'] is None:
        print("WARNING: yu not provided, using yu = 1")
        info['yu'] = 1.0

    # check number of processes
    if info['nprocs'] is None:
        info['nprocs'] = len(args.datafiles)
        print("WARNING: nprocs not provided, using nprocs =", info['nprocs'])

    # check if all the expected input files were provided
    if len(args.datafiles) != info['nprocs']:
        print("ERROR: number of data files (", len(args.datafiles),
              ") does not match number of processes (", info['nprocs'], ")")
        sys.exit()

    if args.debug:
        print('title    = ', info['title'])
        print('varnames = ', info['varnames'])
        print('nprocs   = ', info['nprocs'])
        print('nvar     = ', info['nvar'])
        print('nt       = ', info['nt'])
        print('nx       = ', info['nx'])
        print('xl       = ', info['xl'])
        print('xu       = ', info['xu'])
        print('ny       = ', info['ny'])
        print('yl       = ', info['yl'])
        print('yu       = ', info['yu'])

    return info


# -----------------------------------------------------------------------------
# function to read data file subdomains
# -----------------------------------------------------------------------------


def read_subdomains(args, info):

    import sys
    import shlex
    import numpy as np

    # load subdomain information, store in table
    subdomains = np.zeros((info['nprocs'], 4), dtype=int)

    # get the spatial subdomain owned by each process
    if info['nprocs'] == 1:
        subdomains[0, 0] = 0
        subdomains[0, 1] = info['nx'] - 1
        subdomains[0, 2] = 0
        subdomains[0, 3] = info['ny'] - 1
    else:
        for idx, datafile in enumerate(args.datafiles):

            with open(datafile) as fn:

                # initialize found flags
                found_is = False
                found_ie = False
                found_js = False
                found_je = False

                # read the file line by line
                for line in fn:

                    # skip empty lines
                    if not line.strip():
                        continue

                    # exit after reading initial comment lines
                    if "#" not in line:
                        break

                    # split line into list
                    text = shlex.split(line)

                    # x-direction starting and ending index
                    if "is" in line:
                        subdomains[idx, 0] = int(text[2])
                        found_is = True
                        continue
                    if "ie" in line:
                        subdomains[idx, 1] = int(text[2])
                        found_ie = True
                        continue

                    # y-direction starting and ending index
                    if "js" in line:
                        subdomains[idx, 2] = int(text[2])
                        found_js = True
                        continue
                    if "je" in line:
                        subdomains[idx, 3] = int(text[2])
                        found_je = True
                        continue

                # check if subdomain indices were found
                if not (found_is and found_ie and found_js and found_je):
                    print("ERROR: could not find subdomain indices in",
                          datafile)
                    sys.exit()

    return subdomains


# -----------------------------------------------------------------------------
# read the data files
# -----------------------------------------------------------------------------


def read_data(args, info, subdomains):

    import numpy as np

    # initialize data arrays
    time = np.zeros(info['nt'])
    xvals = np.linspace(info['xl'], info['xu'], info['nx'])
    yvals = np.linspace(info['yl'], info['yu'], info['ny'])
    zdata = np.zeros((info['nt'], info['ny'], info['nx'], info['nvar']))

    # extract data
    for idx, datafile in enumerate(args.datafiles):

        if args.debug:
            print(datafile)

        # load data
        data = np.loadtxt(datafile, dtype=np.double)

        if args.debug:
            print(np.shape(data))

        if np.shape(data)[0] != info['nt']:
            print("WARNING: subdomain", str(idx), "has an incorrect number of"
                  "output times (", np.shape(data)[0], "vs", info['nt'], ")")
            info['nt'] = np.shape(data)[0]

        # x-subdomain indices
        istart = subdomains[idx, 0]
        iend = subdomains[idx, 1]

        # y-subdomain indices
        jstart = subdomains[idx, 2]
        jend = subdomains[idx, 3]

        # local length
        nxl = iend - istart + 1
        nyl = jend - jstart + 1

        if args.debug:
            print(istart, iend, nxl)
            print(jstart, jend, nyl)

        # reshape and save data
        time[:] = data[:, 0]
        for v in range(info['nvar']):
            for i in range(info['nt']):
                zdata[i, jstart:jend+1, istart:iend+1, v] = \
                    np.reshape(data[i, 1+v::info['nvar']], (nyl, nxl))

    return time, xvals, yvals, zdata


# -----------------------------------------------------------------------------
# setup info reused by different plots
# -----------------------------------------------------------------------------


def plot_settings(args, info, time, xvals, yvals, zdata):

    import numpy as np

    # determine extents of plots
    info['zmin'] = np.zeros(info['nvar'])
    info['zmax'] = np.zeros(info['nvar'])

    for v in range(info['nvar']):
        info['zmin'][v] = np.amin(zdata[:, :, :, v])
        info['zmax'][v] = np.amax(zdata[:, :, :, v])

    if args.debug:
        print("z max = ", info['zmax'])
        print("z min = ", info['zmin'])

    # which variables to plot
    if args.plotvars:
        info['pltvars'] = args.plotvars
    else:
        info['pltvars'] = range(info['nvar'])

    # which times to plot
    if args.plottimes:
        info['plttimes'] = args.plottimes
    else:
        info['plttimes'] = range(info['nt'])

    # x-axis label
    if args.xlabel:
        info['xlabel'] = args.xlabel
    else:
        info['xlabel'] = 'x'

    # y-axis label
    if args.ylabel:
        info['ylabel'] = args.ylabel
    else:
        info['ylabel'] = 'y'


# -----------------------------------------------------------------------------
# utility function to combine PDF files
# -----------------------------------------------------------------------------


def merge_pdf(mergefiles, fname):

    import os
    from PyPDF2 import PdfFileMerger

    merger = PdfFileMerger()

    for pdf in mergefiles:
        merger.append(pdf)

    merger.write(fname)
    merger.close()

    for pdf in mergefiles:
        os.remove(pdf)


# -----------------------------------------------------------------------------
# sufrace plot
# -----------------------------------------------------------------------------


def plot_surface(args, info, time, xvals, yvals, zdata):

    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib import cm

    # set x and y meshgrid objects
    X, Y = np.meshgrid(xvals, yvals)

    # generate plots
    for v in info['pltvars']:

        if args.merge:
            mergefiles = list()

        for t in info['plttimes']:

            # create figure and axes
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

            ax.plot_surface(X, Y, zdata[t, :, :, v], rstride=1, cstride=1,
                            cmap=cm.jet, linewidth=0, antialiased=True,
                            shade=True)

            # set axis limits
            ax.set_xlim([info['xl'], info['xu']])
            ax.set_ylim([info['yl'], info['yu']])
            ax.set_zlim(info['zmin'][v], info['zmax'][v])

            # initial perspective
            ax.view_init(20, -120)

            # add axis labels
            plt.xlabel(info['xlabel'])
            plt.ylabel(info['ylabel'])

            # add z-axis label
            if args.zlabel:
                ax.set_zlabel(args.zlabel)
            elif info['varnames']:
                ax.set_zlabel(info['varnames'][v])
            else:
                ax.set_zlabel('z')

            # add title
            tstr = str(time[t])
            if args.title:
                title = args.title
            elif info['title']:
                title = info['title']
            else:
                title = 'Solution'
            plt.title(title + '\nt = ' + tstr)

            # add grid
            if args.grid:
                plt.grid()

            # save plot to file
            if args.save:
                if args.prefix:
                    fname = args.prefix + '_fig_surface_'
                else:
                    fname = 'fig_surface_'
                if info['varnames']:
                    fname += info['varnames'][v]
                else:
                    fname += 'var_' + repr(v).zfill(3)
                fname += '_t_' + repr(t).zfill(3) + '.pdf'
                plt.savefig(fname, bbox_inches='tight')
                if args.merge:
                    mergefiles.append(fname)
            else:
                plt.show()
            plt.close()

        if args.merge:
            if args.prefix:
                fname = args.prefix + '_fig_surface_'
            else:
                fname = 'fig_surface_'
            if info['varnames']:
                fname += info['varnames'][v]
            else:
                fname += 'var_' + repr(v).zfill(3)
            fname += '.pdf'
            merge_pdf(mergefiles, fname)


# -----------------------------------------------------------------------------
# animated sufrace plot
# -----------------------------------------------------------------------------


def plot_surface_ani(args, info, time, xvals, yvals, zdata):

    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation
    from matplotlib import cm

    def update_plot(frame_number, zarray, v, plot):
        plot[0].remove()
        plot[0] = ax.plot_surface(X, Y, zarray[frame_number, :, :, v],
                                  cmap=cm.jet)

        tstr = str(time[frame_number])
        if args.title:
            title = args.title
        elif info['title']:
            title = info['title']
        else:
            title = 'Solution'
        plt.title(title + '\nt = ' + tstr)

        return plot,

    # set x and y meshgrid objects
    X, Y = np.meshgrid(xvals, yvals)

    # generate plots
    for v in info['pltvars']:

        # create figure and axes
        fig = plt.figure()
        ax = plt.axes(projection='3d')

        plot = [ax.plot_surface(X, Y, zdata[0, :, :, v], rstride=1, cstride=1,
                                cmap=cm.jet, linewidth=0, antialiased=True,
                                shade=True)]

        # set axis limits
        ax.set_xlim([info['xl'], info['xu']])
        ax.set_ylim([info['yl'], info['yu']])
        ax.set_zlim([info['zmin'][v], info['zmax'][v]])

        # initial perspective
        ax.view_init(20, -120)

        # add x-axis label
        if args.xlabel:
            plt.xlabel(args.xlabel)
        else:
            ax.set_xlabel('x')

        # add y-axis label
        if args.ylabel:
            plt.ylabel(args.ylabel)
        else:
            ax.set_ylabel('y')

        # add z-axis label
        if args.zlabel:
            ax.set_zlabel(args.zlabel)
        elif info['varnames']:
            ax.set_zlabel(info['varnames'][v])
        else:
            ax.set_zlabel('z')

        # add grid
        if args.grid:
            plt.grid()

        fps = 2          # frame per sec
        frn = len(time)  # number of frames in the animation

        # create animation
        ani = animation.FuncAnimation(fig, update_plot, frn,
                                      fargs=(zdata, v, plot),
                                      interval=1000/fps)

        # save animation to file
        if args.save:
            if args.prefix:
                fname = args.prefix + '_ani_surface_'
            else:
                fname = 'ani_surface_'
            if info['varnames']:
                fname += info['varnames'][v]
            else:
                fname += 'var_' + repr(v).zfill(3)
            ani.save(fname + '.mp4', dpi=200, fps=fps)
        else:
            plt.show()
        plt.close()


# -----------------------------------------------------------------------------
# contour plot
# -----------------------------------------------------------------------------


def plot_contour(args, info, time, xvals, yvals, zdata):

    import numpy as np
    import matplotlib.pyplot as plt

    # set x and y meshgrid objects
    X, Y = np.meshgrid(xvals, yvals)

    # generate plots
    for v in info['pltvars']:

        levels = np.linspace(info['zmin'][v], info['zmax'][v], 100)
        ticks = np.linspace(info['zmin'][v], info['zmax'][v], 10)

        for t in info['plttimes']:

            # create figure and axes
            fig, ax = plt.subplots()

            cf = ax.contourf(X, Y, zdata[t, :, :, v], levels=levels,
                             cmap="coolwarm", extend="both")
            fig.colorbar(cf, ax=ax, fraction=0.046, pad=0.04, ticks=ticks)

            # set axis limits
            ax.set_xlim([info['xl'], info['xu']])
            ax.set_ylim([info['yl'], info['yu']])

            # add axis labels
            plt.xlabel(info['xlabel'])
            plt.ylabel(info['ylabel'])

            # add title
            tstr = str(time[t])
            if args.title:
                plt.title(args.title + ' at t = ' + tstr)
            elif info['title']:
                plt.title(info['title'] + ' at t = ' + tstr)
            else:
                plt.title('Solution at t = ' + tstr)

            # add grid
            if args.grid:
                plt.grid()

            # save plot to file
            if args.save:
                if args.prefix:
                    fname = args.prefix + '_fig_contour_'
                else:
                    fname = 'fig_contour_'
                if info['varnames']:
                    fname += info['varnames'][v]
                else:
                    fname += 'var_' + repr(v).zfill(3)
                fname += '_t_' + repr(t).zfill(3) + '.pdf'
                plt.savefig(fname, bbox_inches='tight')
            else:
                plt.show()
            plt.close()


# -----------------------------------------------------------------------------
# animated contour plot
# -----------------------------------------------------------------------------


def plot_contour_ani(args, info, time, xvals, yvals, zdata):

    import numpy as np
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation

    def update_plot(frame_number, zarray, v, plot):
        plot[0] = ax.contourf(X, Y, zdata[frame_number, :, :, v],
                              levels=levels, cmap="coolwarm", extend="both")

        tstr = str(time[frame_number])
        if args.title:
            title = args.title
        elif info['title']:
            title = info['title']
        else:
            title = 'Solution'
        plt.title(title + '\nt = ' + tstr)

        return plot,

    # set x and y meshgrid objects
    X, Y = np.meshgrid(xvals, yvals)

    # generate plots
    for v in info['pltvars']:

        levels = np.linspace(info['zmin'][v], info['zmax'][v], 100)
        ticks = np.linspace(info['zmin'][v], info['zmax'][v], 10)

        # create figure and axes
        fig, ax = plt.subplots()

        plot = [ax.contourf(X, Y, zdata[0, :, :, v], levels=levels,
                            cmap="coolwarm", extend="both")]
        fig.colorbar(plot[0], ax=ax, fraction=0.046, pad=0.04, ticks=ticks)

        # set axis limits
        ax.set_xlim([info['xl'], info['xu']])
        ax.set_ylim([info['yl'], info['yu']])

        # add axis labels
        plt.xlabel(info['xlabel'])
        plt.ylabel(info['ylabel'])

        # add grid
        if args.grid:
            plt.grid()

        fps = 2          # frame per sec
        frn = len(time)  # number of frames in the animation

        # create animation
        ani = animation.FuncAnimation(fig, update_plot, frn,
                                      fargs=(zdata, v, plot),
                                      interval=1000/fps)

        # save animation to file
        if args.save:
            if args.prefix:
                fname = args.prefix + '_ani_contour_'
            else:
                fname = 'ani_contour_'
            if info['varnames']:
                fname += info['varnames'][v]
            else:
                fname += 'var_' + repr(v).zfill(3)
            ani.save(fname + '.mp4', dpi=200, fps=fps)
        else:
            plt.show()
        plt.close()


# -----------------------------------------------------------------------------
# 1d slice evolution with new figure for each variable
# -----------------------------------------------------------------------------


def plot_slice_vars(args, info, time, svals, sdata, hlabel, suffix):

    import numpy as np
    import matplotlib.pyplot as plt

    # determine extents of slice plot
    smin = np.zeros(info['nvar'])
    smax = np.zeros(info['nvar'])

    for v in range(info['nvar']):
        smin[v] = np.amin(sdata[:, :, v])
        smax[v] = np.amax(sdata[:, :, v])

    if args.debug:
        print(smin)
        print(smax)

    # set labels for the plot legend
    if args.labels:
        label = args.labels
    else:
        label = ["%.2f" % t for t in time]

    # create plot for each variable
    for v in info['pltvars']:

        # create figure and axes
        fig, ax = plt.subplots()

        # add each output time to the plot
        for t in info['plttimes']:
            ax.plot(svals, sdata[t, :, v], label=label[t])

        # set axis limits
        ax.set_xlim([svals[0], svals[-1]])
        ax.set_ylim([0.99 * smin[v], 1.01 * smax[v]])

        # add legend
        ax.legend(title="Times", bbox_to_anchor=(1.02, 1), loc="upper left")

        # add x-axis label
        ax.set_xlabel(hlabel)

        # add y-axis label
        if args.zlabel:
            ax.set_ylabel(args.zlabel)
        else:
            if info['varnames']:
                ax.set_ylabel(info['varnames'][v])
            else:
                ax.set_ylabel('variable ' + repr(v))

        # add title
        if args.title:
            plt.title(args.title + suffix)
        elif info['title']:
            plt.title(info['title'] + suffix)
        else:
            if info['varnames']:
                plt.title("Evolution of " + info['varnames'][v] + suffix)
            else:
                plt.title("Evolution of variable " + repr(v) + suffix)

        # add grid
        if args.grid:
            plt.grid()

        # save plot to file
        if args.save:
            if args.prefix:
                fname = args.prefix + '_fig_slice_'
            else:
                fname = 'fig_slice_'
            if info['varnames']:
                fname += info['varnames'][v]
            else:
                fname += 'var_' + repr(v).zfill(3)
            plt.savefig(fname + '.pdf', bbox_inches='tight')
        else:
            plt.show()
        plt.close()


# -----------------------------------------------------------------------------
# 1d slice evolution with new figure for each time
# -----------------------------------------------------------------------------


def plot_slice_time(args, info, time, svals, sdata, hlabel, suffix):

    import numpy as np
    import matplotlib.pyplot as plt

    # determine extents of slice plot
    smin = np.amin(sdata)
    smax = np.amax(sdata)

    if args.debug:
        print(smin)
        print(smax)

    # set labels for the plot legend
    if args.labels:
        label = args.labels
    elif info['varnames']:
        label = info['varnames']
    else:
        label = [None] * info['nvar']

    # create plot for each variable
    for t in info['plttimes']:

        # create figure and axes
        fig, ax = plt.subplots()

        # add each output time to the plot
        for v in info['pltvars']:
            ax.plot(svals, sdata[t, :, v], label=label[v])

        # set axis limits
        ax.set_xlim([svals[0], svals[-1]])
        ax.set_ylim([0.99 * smin, 1.01 * smax])

        # add legend
        ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

        # add x-axis label
        ax.set_xlabel(hlabel)

        # add y-axis label
        if args.zlabel:
            ax.set_ylabel(args.zlabel)

        # add title
        tstr = str(time[t])
        if args.title:
            plt.title(args.title + suffix + ' and t = ' + tstr)
        elif info['title']:
            plt.title(info['title'] + suffix + ' and t = ' + tstr)
        else:
            plt.title("Evolution" + suffix + ' and t = ' + tstr)

        # add grid
        if args.grid:
            plt.grid()

        # save plot to file
        if args.save:
            if args.prefix:
                fname = args.prefix + '_fig_slice_t_'
            else:
                fname = 'fig_slice_t_'
            fname += repr(t).zfill(3) + '.pdf'
            plt.savefig(fname, bbox_inches='tight')
        else:
            plt.show()
        plt.close()


# -----------------------------------------------------------------------------
# point evolution
# -----------------------------------------------------------------------------


def plot_point(args, info, time, pdata, suffix):

    import matplotlib.pyplot as plt

    # set labels for the plot legend
    if args.labels:
        label = args.labels
    elif info['varnames']:
        label = info['varnames']
    else:
        label = [None] * info['nvar']

    # create figure and axes
    fig, ax = plt.subplots()

    # create plot for each variable
    for v in info['pltvars']:
        ax.plot(time, pdata[:, v], label=label[v])

    # add legend
    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    # add x-axis label
    ax.set_xlabel("time")

    # add title
    if args.title:
        plt.title(args.title + suffix)
    elif info['title']:
        plt.title(info['title'] + suffix)
    else:
        plt.title("Evolution" + suffix)

    # add grid
    if args.grid:
        plt.grid()

    # save plot to file
    if args.save:
        if args.prefix:
            fname = args.prefix + '_fig_point'
        else:
            fname = 'fig_point'
        plt.savefig(fname + '.pdf', bbox_inches='tight')
    else:
        plt.show()
    plt.close()


# -----------------------------------------------------------------------------
# run the main routine
# -----------------------------------------------------------------------------


if __name__ == '__main__':
    import sys
    sys.exit(main())



# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

import sys, os
sys.path.append(os.path.dirname(os.path.abspath('../../../shared/versions.py')))
from versions import *

# -- Create new object types --------------------------------------------------

from sphinx.application import Sphinx

def setup(app: Sphinx):
    app.add_object_type('cmakeoption', 'cmakeop', 'single: %s (CMake option)')
    app.add_config_value('package_name', '', 'env', types=[str])

# -- General configuration ----------------------------------------------------

# Set variable used to determine which package documentation this is
# Can be one of 'arkode', 'cvode', 'cvodes', 'ida', 'idas', 'kinsol' or 'super'
package_name = 'arkode'

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '4.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx_rtd_theme', 'sphinx.ext.ifconfig', 'sphinx.ext.mathjax',
              'sphinxfortran.fortran_domain', 'sphinxcontrib.bibtex', 'sphinx_copybutton']

# References
bibtex_bibfiles = ['../../../shared/sundials.bib']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'User Documentation for ARKODE'
copyright = u"""2012-{year}, Daniel R. Reynolds, David J. Gardner, Carol S. Woodward, Rujeko Chinomona, and Cody J. Balos""".format(year = year)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '{arkode_version}'.format(arkode_version=arkode_version)
sun_version = '{sundials_version}'.format(sundials_version=sundials_version)

# Set the date format (full-month-name day, full-year)
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'
highlight_language = "c"

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# Number figures, tables, and code blocks (can reference by number with numref)
numfig = True

# Override format strings that numref/numfig uses
numfig_format = {
  'section': '§%s'
}

rst_prolog = open('../../../shared/global.rst.txt', 'r').read()

rst_epilog = """
.. |YEAR| replace:: {year}
.. |CVODE_VERSION| replace:: {cvode_version}
.. |CVODES_VERSION| replace:: {cvodes_version}
.. |ARKODE_VERSION| replace:: {arkode_version}
.. |IDA_VERSION| replace:: {ida_version}
.. |IDAS_VERSION| replace:: {idas_version}
.. |KINSOL_VERSION| replace:: {kinsol_version}
""".format(year = year,
cvode_version = cvode_version,
cvodes_version = cvodes_version,
arkode_version = arkode_version,
ida_version = ida_version,
idas_version = idas_version,
kinsol_version = kinsol_version
)

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Set theme options
html_theme_options = {
    # Allow unlimited depth in table of contents tree
    'navigation_depth': -1
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'figs/sundials_logo_blue.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['../../../shared/_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
  'css/custom.css'
]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'ARKODEdoc'


# -- Options for LaTeX output --------------------------------------------------

latex_additional_files = ['../../../shared/latex/preamble.tex.txt',
                          '../../../shared/latex/cover_pages.tex.txt']

im_number = "LLNL-SM-668082"

# 1. the rst file name used to generate the LaTeX file
# 2. the name of the LaTeX file to generate (and the resulting PDF file name)
# 3. the document title
# 4. text for \author
# 5. the LaTeX theme
# 6. include the file from 1. in the output
tex_author = r'''
    Daniel R. Reynolds$^1$,
    David J. Gardner$^2$,
    Carol S. Woodward$^2$,
    Rujeko Chinomona$^3$, and
    Cody J. Balos$^2$ \\
    \\
    {\em $^1$Department of Mathematics, Southern Methodist University} \\
    {\em $^2$Center for Applied Scientific Computing, Lawrence Livermore National Laboratory} \\
    {\em $^3$Department of Mathematics, Temple University}
    '''

latex_documents = [('index', 'ark_guide.tex', project,
                    tex_author, 'manual', False)]

# make sure the doc logo gets copied to the build directory
latex_logo = 'figs/doc_logo_blue.pdf'

# LaTeX customizations
latex_elements = {
# paper size option of the document class
'papersize': 'letterpaper',
# font size option of the document class
'pointsize': '10pt',
# set the version number/release name
'releasename' : version,
# arguments to the sphinxsetup macro
'sphinxsetup':
    # the color for titles
    'TitleColor={RGB}{0,0,0},'+
    # disable frames around code-blocks
    'verbatimwithframe=false,'+
    # do not wrap long lines in code-blocks
    'verbatimwrapslines=false,'+
    # background color for code-blocks
    'VerbatimColor={RGB}{240.0,240.0,240.0},'+
    # font used by heading
    'HeaderFamily=\\rmfamily\\bfseries,' +
    # line breaks are allowed inside inline literals
    'inlineliteralwraps=true',
# disable the fncychap package
'fncychap':'',
# figure alignment options
'figure_align': 'htbp',
# additional preamble content
'preamble':
    '\\input{preamble.tex.txt}\n'+
    '\\newcommand{\\sunreleasename}{' + sun_version + '}\n' +
    '\\newcommand{\\imnumber}{' + im_number + '}\n',
# extra class options
'extraclassoptions': 'twoside,openright',
# custom maketitle
'maketitle': '\\input{cover_pages.tex.txt}'
}


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'ARKODE', u'ARKODE Documentation',
     [u'Daniel R. Reynolds, David J. Gardner, Carol S. Woodward, Rujeko Chinomona, and Cody J. Balos'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'ARKODE', u'ARKODE Documentation',
   u'Daniel R. Reynolds, David J. Gardner, Carol S. Woodward, Rujeko Chinomona, and Cody J. Balos', 'ARKODE',
   'Time integration package for multi-rate systems of ordinary differntial equations.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Test all installed Makefiles or CMakeLists.txt files under a given directory
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# main routine
# -----------------------------------------------------------------------------

def main():

    import argparse
    import fnmatch
    import os
    import re
    import subprocess

    parser = argparse.ArgumentParser(
        description='Find and build installed examples',
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('directory', type=str,
                        help='Directory to search for build files')

    parser.add_argument('--cmake', action='store_true',
                        help='CMake build')

    parser.add_argument('--test', action='store_true',
                        help='Test builds')

    parser.add_argument('--clean', action='store_true',
                        help='Clean builds')

    parser.add_argument('--regex', type=str,
                        help='Regular expression for filtering example directories')

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help='Verbose output')

    parser.add_argument('--failfast', action='store_true',
                        help='Stop on first failure')

    # parse command line args
    args = parser.parse_args()

    # get current directory
    cwd = os.getcwd()

    # which build file to search for
    if args.cmake:
        filename = "CMakeLists.txt"
    else:
        filename = "Makefile"

    # get list of build files
    buildfiles = []
    for root, dirs, files in os.walk(args.directory):
        for fn in fnmatch.filter(files, filename):
            buildfiles.append(os.path.join(root, fn))

    # output list of build files
    if args.verbose > 0:
        print(f"Total files: {len(buildfiles)}")
    if args.verbose > 2:
        for bf in buildfiles:
            print(bf)

    # filter files
    if args.regex:
        regex = re.compile(args.regex)
        buildfiles = [ bf for bf in buildfiles if re.search(regex, bf) ]
        if args.verbose > 0:
            print(f"Total files (filtered): {len(buildfiles)}")
        if args.verbose > 2:
            for bf in buildfiles:
                print(bf)

    # total files to test
    total = len(buildfiles)

    # build examples
    fail = False
    errors = []
    for i, bf in enumerate(buildfiles):

        print(f"[{i+1} of {total}] Building: {bf}")

        # move to example directory
        os.chdir(os.path.dirname(bf))

        # clean and move on
        if args.clean:
            ret = subprocess.call('make clean', shell=True,
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)
            # return to original directory
            os.chdir(cwd)
            continue

        # confgure cmake if necessary
        configfail = False
        if args.cmake:
            if os.path.isfile('Makefile'):
                os.remove('Makefile')
            ret = subprocess.call('cmake -DCMAKE_VERBOSE_MAKEFILE=ON .',
                                  shell=True, stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)
            if args.verbose > 0:
                print(f"  Config return: {ret}")
            if ret != 0:
                errors.append(os.path.dirname(bf))
                configfail = True

        # make examples
        buildfail = False
        if not configfail:
            ret = subprocess.call('make', shell=True,
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)
            if args.verbose > 0:
                print(f"  Build return: {ret}")
            if ret != 0:
                errors.append(os.path.dirname(bf))
                buildfail = True

        # test examples
        testfail = False
        if not configfail and not buildfail and args.test:
            ret = subprocess.call('make test', shell=True,
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)
            if args.verbose > 0:
                print(f"  Test return: {ret}")
            if ret != 0:
                errors.append(os.path.dirname(bf))
                testfail = True

        # return to original directory
        os.chdir(cwd)

        # set test status
        if configfail or buildfail or testfail:
            fail = True

        # stop on first error
        if args.failfast and fail:
            break

    # print list of failed builds
    if fail:
        print("The following examples failed:")
        for err in errors:
            print(err)
        sys.exit(1)
    else:
        print("All builds successful.")

# -----------------------------------------------------------------------------
# run the main routine
# -----------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    sys.exit(main())



# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
sundials_version = 'v6.4.1'
arkode_version = 'v5.4.1'
cvode_version = 'v6.4.1'
cvodes_version = 'v6.4.1'
ida_version = 'v6.4.1'
idas_version = 'v5.4.1'
kinsol_version = 'v6.4.1'
year = '2022'



#!/usr/bin/env python
# ------------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL 
# ------------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ------------------------------------------------------------------------------
# Find unique lines from a file containing a specific string and write them to
# a separate file.
#
# Example usage:
#   Find unique compiler warnings in make.log and write them to warnings.txt:
#   $ ./findlines.py make.log warning warnings.txt
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# main routine
# ------------------------------------------------------------------------------
def main():

    import argparse
    import sys, os
    import shlex

    parser = argparse.ArgumentParser(
        description='Find unique lines in <readfile> containing <key> '+
        'and write them to <outfile>',
        formatter_class=argparse.RawTextHelpFormatter)
    
    parser.add_argument('readfile',type=str,
                        help='File to search')

    parser.add_argument('key',type=str,
                        help='String to find in lines of <readfile>')

    parser.add_argument('outfile',type=str,
                        help='Output file for lines in <readfile> containing <key>')

    # parse command line args 
    args = parser.parse_args()
   
    # check if readfile exists
    if (not os.path.isfile(args.readfile)):
        print "Error:",args.readfile,"not found"
        return -1

    # check that outfile does not already exist
    if (os.path.isfile(args.outfile)):
        print "Error:",args.outfile,"already exists"
        return -1

    # create set to hold lines already seen
    lines_seen = set() 

    with open(args.readfile,"r") as fr:
        with open(args.outfile,"w") as fw:
            # loop over every line in readfile
            for line in fr:
                # check if line contains the key
                if (args.key in line):
                    # if not a duplicate line then 
                    # add the line to the set and 
                    # output it to the results file
                    if line not in lines_seen: 
                        lines_seen.add(line)
                        print >> fw, line

# ------------------------------------------------------------------------------
# run the main routine
# ------------------------------------------------------------------------------
if __name__ == '__main__':
    import sys
    sys.exit(main())



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

# imports
import glob
import sys
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


# load pickled data
def load_data(file):
    data = np.load(file)
    m = data['mesh']
    t = data['t']
    u = data['u']
    v = data['v']
    w = data['w']

    hx = m[0,1] - m[0,0]
    hy = m[1,1] - m[1,0]
    hz = m[2,1] - m[2,0]

    return { 'm': m, 'h': (hx,hy,hz), 't': t, 'u': u, 'v': v, 'w': w }


# grid function norm
def norm_3Dgrid(h, x, q=1):
    hx,hy,hz = h
    return (hx*hy*hz*np.sum(np.abs(x)**q, axis=(1,2,3)))**(1/q)


# computer order of accuracy p
def calc_order(h1, Eh1, h2, Eh2):
    return np.log( Eh1/Eh2 ) / np.log( np.prod(h1)/np.prod(h2) )


# load data files
h_over_8 = load_data('middle-h/output-with-h-1.04e-02.npz')
h_over_4 = load_data('large-h/output-with-h-2.08e-02.npz')
# h_over_2 = load_data('larger-h/output-with-h-4.16e-02.npz')
h_over_1 = load_data('largest-h/output-with-h-8.33e-02.npz')

for component in ['u', 'v', 'w']:
    # Restrict reference to the coarsest grid
    ref = h_over_8[component][:,::8,::8,::8]

    # Now compute E(h) = ||U(h) - \bar{U}(h)|| using the grid-function norm
    Eh_over_4 = norm_3Dgrid(h_over_4['h'], h_over_4[component][:,::4,::4,::4] - ref)
    Eh_over_1 = norm_3Dgrid(h_over_1['h'], h_over_1[component][:,:,:,:] - ref)

    # Compute order p as in O(h^p)
    p = calc_order(h_over_1['h'], Eh_over_1, h_over_4['h'], Eh_over_4)
    print('min p for %s component: %.4f' % (component, np.min(p)))

    # Plot error across time
    plt.figure()
    plt.plot(h_over_8['t'], Eh_over_4, 'r-')
    plt.plot(h_over_8['t'], Eh_over_1, 'b-')
    plt.ylabel('||E(hx,hy,hz)||')
    plt.xlabel('time')
    plt.savefig('error-in-time-plot-%s.png' % component)

    # Plot error norm with respect to h
    plt.figure()
    x = np.array([np.prod(h_over_4['h']), np.prod(h_over_1['h'])])
    plt.plot(x, x, 'k-')
    plt.plot(x, x**2, 'k-')
    plt.plot(x, [np.linalg.norm(Eh_over_4, np.Inf), np.linalg.norm(Eh_over_1, np.Inf)], 'r-')
    plt.legend(['1st order', '2nd order', 'actual'])
    plt.ylabel('|| ||E(hx,hy,hz)|| ||_inf')
    plt.xlabel('hx * hy * hz')
    plt.yscale('log')
    plt.xscale('log')
    plt.savefig('error-plot-%s.png' % component)



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# This script plots the output from test_nector_performance_* and assumes:
#   1. vector lengths are powers of two starting from 0, and
#   2. output files are named: output_nelem_nvec_nsum_ntest_timing.txt
# where nelem is the number of elements in the vector, nvec is the nuber of
# vectors, nsum is the number of sums, ntest is the number of tests, and timing
# indicates if timing was enabled.
# -----------------------------------------------------------------------------

def main():

    import argparse
    import os, sys
    import shlex
    import glob

    import numpy as np
    import scipy.stats as st

    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.ticker as mtick

    parser = argparse.ArgumentParser(
        description='Plot data from NVector performance tests')

    parser.add_argument('op', type=str,
                        help='Which NVector operation to plot')

    parser.add_argument('datadir', type=str,
                        help='Directory where test output files are located')

    parser.add_argument('--noplots', dest='noplots', action='store_true',
                        help='Turn on plots for time vs number of elements')

    parser.add_argument('--logx', dest='logx', action='store_true',
                        help='Generate plots for speedup with log scale for the x axis (number of elements')

    parser.add_argument('--fused', dest='fused', action='store_true',
                        help='Operation is a fused op')

    parser.add_argument('--show', dest='show', action='store_true',
                        help='Display plots rather than saving to file')

    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Turn on debugging output')

    # parse command line args
    args = parser.parse_args()

    if (args.debug):
        print(args)

    # check for test data directory
    if (not os.path.isdir(args.datadir)):
        print("ERROR:",args.datadir,"does not exist")
        sys.exit()

    # sort output files
    output_baseline = sorted(glob.glob(args.datadir+'/output*-old.log'))
    output_new      = sorted(glob.glob(args.datadir+'/output*-new.log'))
    output = output_baseline + output_new

    if (args.debug):
        print("output files")
        print(len(output))
        for i in range(len(output)):
            print(output[i])

    # figure out vector sizes, number of vectors, and number of sums
    nelem = []
    nvec  = []
    nsum  = []
    ntest = []

    # parse file names to get input parameters
    for f in output:

        split_fout = f.split("/")[-1]
        split_fout = split_fout.split("_")

        ne = int(split_fout[1])
        nv = int(split_fout[2])
        ns = int(split_fout[3])
        nt = int(split_fout[4])

        if (not ne in nelem):
            nelem.append(ne)

        if (not nv in nvec):
            nvec.append(nv)

        if (not ns in nsum):
            nsum.append(ns)

        if (not nt in ntest):
            ntest.append(nt)

    if (len(ntest) != 1):
        print("Warning: Unequal numbers of tests")

    nelem.sort()

    if (args.debug):
        print("nelem:",nelem, len(nelem))
        print("nvec: ",nvec,  len(nvec))
        print("nsum: ",nsum,  len(nsum))
        print("ntest:",ntest, len(ntest))

    # allocate numpy arrays for timing data
    avg_denom  = np.zeros([len(nvec), len(nelem)])
    sdev_denom = np.zeros([len(nvec), len(nelem)])
    avg_numer  = np.zeros([len(nvec), len(nelem)])
    sdev_numer = np.zeros([len(nvec), len(nelem)])
    avg_ratio = np.zeros([len(nvec), len(nelem)])

    # read 'baseline' files
    for f in output_baseline:
        if (args.debug):
            print("Reading:",f)
        # get test inputs from file name
        split_fout = f.split("/")[-1]
        split_fout = split_fout.split("_")
        ne = int(split_fout[1])
        nv = int(split_fout[2])
        ns = int(split_fout[3])
        with open(f) as fout:
            for line in fout:
                # split line into list
                split_line = shlex.split(line)
                # skip blank lines
                if (not split_line):
                    continue
                # tests finished, stop reading file
                if (split_line[0] == "Finished"):
                    break
                # check if the operation is the one we want and get data
                if (args.op == split_line[0]):
                    i = nvec.index(nv)
                    j = nelem.index(ne)
                    avg_numer[i][j]  = float(split_line[1])
                    sdev_numer[i][j] = float(split_line[2])

    # read output files
    for f in output_new:
        if (args.debug):
            print("Reading:",f)
        # get test inputs from file name
        split_fout = f.split("/")[-1]
        split_fout = split_fout.split("_")
        ne = int(split_fout[1])
        nv = int(split_fout[2])
        ns = int(split_fout[3])
        with open(f) as fout:
            for line in fout:
                # split line into list
                split_line = shlex.split(line)
                # skip blank lines
                if (not split_line):
                    continue
                # tests finished, stop reading file
                if (split_line[0] == "Finished"):
                    break
                # check if the operation is the one we want and get data
                if (args.op == split_line[0]):
                    i = nvec.index(nv)
                    j = nelem.index(ne)
                    avg_denom[i][j]  = float(split_line[1])
                    sdev_denom[i][j] = float(split_line[2])
                    avg_ratio[i][j] = avg_numer[i][j] / avg_denom[i][j]

    # --------------------------------------------------------------------------
    # Confidence Interval
    # --------------------------------------------------------------------------

    # allocate arrays for the upper and lower bounds of the confidence interval
    lower_denom   = np.zeros([len(nvec), len(nelem)])
    upper_denom   = np.zeros([len(nvec), len(nelem)])
    lower_numer = np.zeros([len(nvec), len(nelem)])
    upper_numer = np.zeros([len(nvec), len(nelem)])

    # critical value for 99% confidence interval
    if (ntest[0] < 30):
        # student's t distribution
        cv = st.t.interval(0.99, ntest[0]-1)[1]
    else:
        # normal distribution
        cv = st.norm.ppf(0.995)

    # confidence intervals
    cdev_denom  = cv * sdev_denom / np.sqrt(ntest[0])
    lower_denom = avg_denom - cdev_denom
    upper_denom = avg_denom + cdev_denom

    cdev_numer  = cv * sdev_numer / np.sqrt(ntest[0])
    lower_numer = avg_numer - cdev_numer
    upper_numer = avg_numer + cdev_numer

    # check if the new average times are within the baseline confidence interval
    denom_in = np.where(np.logical_and(avg_denom < upper_numer,
                                       avg_denom > lower_numer))

    # check if the baseline average times are within the new confidence interval
    numer_in = np.where(np.logical_and(avg_numer < upper_denom,
                                         avg_numer > lower_denom))

    # get which numbers of vectors and elements for new tests are in the
    # confidence interval of the baseline times
    df = np.zeros([len(denom_in[0])])
    vf = np.zeros([len(denom_in[0])])
    ef = np.zeros([len(denom_in[0])])

    for i in range(len(denom_in[0])):
        vf[i] = nvec[denom_in[0][i]]
        ef[i] = np.log2(nelem[denom_in[1][i]])
        df[i] = 1

    if (args.debug):
        print('vf:', vf)
        print('ef:', ef)

    # get which numbers of vectors and elements for baseline tests are in the
    # confidence interval of the new times
    du = np.zeros([len(numer_in[0])])
    vu = np.zeros([len(numer_in[0])])
    eu = np.zeros([len(numer_in[0])])

    for i in range(len(numer_in[0])):
        vu[i] = nvec[numer_in[0][i]]
        eu[i] = np.log2(nelem[numer_in[1][i]])
        du[i] = 1

    if (args.debug):
        print('vu:', vu)
        print('eu:', eu)

    # --------------------------------------------------------------------------
    # Output ratios
    # --------------------------------------------------------------------------

    # output the matrix of ratios in reverse row order to match the heatmap
    print(args.op)

    print("avg. new")
    for i in reversed(range(len(nvec))):
        print('%2d' % int(i+1), str(avg_denom[i]).replace('\n', ''))
    print()

    print("avg. baseline")
    for i in reversed(range(len(nvec))):
        print('%2d' % int(i+1), str(avg_numer[i]).replace('\n', ''))
    print()

    print("avg. ratio (speedup)")
    for i in reversed(range(len(nvec))):
        print('%2d' % int(i+1), str(avg_ratio[i]).replace('\n', ''))
    print()

    # --------------------------------------------------------------------------
    # Speedup v. Number of Elements Plots
    # --------------------------------------------------------------------------
    if (not args.noplots):

        colors = ['#000000','#a6cee3','#1f78b4','#b2df8a','#33a02c',
                  '#fb9a99','#e31a1c','#fdbf6f','#ff7f00','#cab2d6',
                  '#6a3d9a','#ffff99','#b15928']

        hatch = [ '/','\\','-','+','x','o','O','.','*']

        # --------------------------------------------------------------------------
        # Combined Number of Vectors Plots
        # --------------------------------------------------------------------------
        fig = plt.figure()
        ax = fig.add_subplot(111)

        if args.fused:
            indices = range(0,len(nvec))
        else:
            indices = range(len(nvec)-1,len(nvec))

        for i in indices:
            lab = 'num. vecs %d' % nvec[i]
            if (args.logx):
                ax.plot(nelem, avg_ratio[i],
                        color=colors[i], linestyle='-', label=lab)
                ax.set_xscale('log')
            else:
                ax.plot(nelem, avg_ratio[i],
                        color=colors[i], linestyle='-', label=lab)
                # # plot confidence interval
                # ax.fill_between(nelem, lower_denom[i], upper_denom[i],
                #                 color=colors[i], alpha=0.3)
                # ax.fill_between(nelem, lower_numer[i], upper_numer[i],
                #                 color=colors[i], hatch='.', alpha=0.3)

        ax.legend()
        ax.grid()

        plt.title('Average Speedup \n'+args.op)
        plt.xlabel('vector length')
        plt.ylabel('speedup (baseline/new)')

        if (args.show):
            plt.show()
        else:
            if (args.logx):
                fname=args.op+'-nvec-all-logx.pdf'
            else:
                fname=args.op+'-nvec-all.pdf'
                plt.ticklabel_format(axis='both',style='sci')
            plt.savefig(fname)
        plt.close()


# ===============================================================================

if __name__ == "__main__":
    main()

# EOF



#!/usr/bin/env python
# ----------------------------------------------------------------
# Programmer(s): Daniel R. Reynolds @ SMU
# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2022, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
# matplotlib-based plotting script for heat2D.cpp example

# imports
import sys
import numpy as np
from pylab import *
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.pyplot as plt

# determine the number of MPI processes used
nprocs=1
for i in range(1000):
    sname = 'heat2d_subdomain.' + repr(i).zfill(3) + '.txt'
    try:
        f = open(sname,'r')
        f.close()
    except IOError:
        nprocs = i
        break

# load subdomain information, store in table
subdomains = np.zeros((nprocs,4), dtype=np.int)
for i in range(nprocs):
    sname = 'heat2d_subdomain.' + repr(i).zfill(3) + '.txt'
    subd = np.loadtxt(sname, dtype=np.int)
    if (i == 0):
        nx = subd[0]
        ny = subd[1]
    else:
        if ((subd[0] != nx) or (subd[1] != ny)):
            sys.exit("error: subdomain files incompatible (clean up and re-run test)")
    subdomains[i,:] = subd[2:6]
    
# load first processor's data, and determine total number of time steps
data = np.loadtxt('heat2d.000.txt', dtype=np.double)
nt = np.shape(data)[0]

# create empty array for all solution data
results = np.zeros((nt,ny,nx))

# insert first processor's data into results array
istart = subdomains[0,0]-1
iend = subdomains[0,1]-1
jstart = subdomains[0,2]-1
jend = subdomains[0,3]-1
nxl = iend-istart+1
nyl = jend-jstart+1
for i in range(nt):
    results[i,jstart:jend+1,istart:iend+1] = np.reshape(data[i,:], (nyl,nxl))
    
# iterate over remaining data files, inserting into output
if (nprocs > 1):
    for isub in range(1,nprocs):
        data = np.loadtxt('heat2d.' + repr(isub).zfill(3) + '.txt', dtype=np.double)
        # check that subdomain has correct number of time steps
        if (np.shape(data)[0] != nt):
            sys.exit('error: subdomain ' + isub + ' has an incorrect number of time steps')
        istart = subdomains[isub,0]-1
        iend = subdomains[isub,1]-1
        jstart = subdomains[isub,2]-1
        jend = subdomains[isub,3]-1
        nxl = iend-istart+1
        nyl = jend-jstart+1
        for i in range(nt):
            results[i,jstart:jend+1,istart:iend+1] = np.reshape(data[i,:], (nyl,nxl))

# determine extents of plots
maxtemp = 1.1*results.max()
mintemp = 0.9*results.min()

# generate plots of results
kx = 0.5
ky = 0.75
kprod = (kx+4.0*ky)*np.pi**2
dt = 0.015
for tstep in range(nt):

    # set string constants for output plots, current time, mesh size
    pname = 'heat2d_surf.' + repr(tstep).zfill(3) + '.png'
    cname = 'heat2d_err.' + repr(tstep).zfill(3) + '.png'
    tstr  = repr(tstep)
    nxstr = repr(nx)
    nystr = repr(ny)

    # set x and y meshgrid objects
    xspan = np.linspace(0.0, 1.0, nx)
    yspan = np.linspace(0.0, 1.0, ny)
    X,Y = np.meshgrid(xspan,yspan)

    # plot current solution as a surface, and save to disk
    fig = plt.figure(1)
    ax = fig.add_subplot(111, projection='3d')
    ax.plot_surface(X, Y, results[tstep,:,:], rstride=1, cstride=1, 
                    cmap=cm.jet, linewidth=0, antialiased=True, shade=True)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlim((mintemp, maxtemp))
    ax.view_init(20,45)
    title('u(x,y) at output ' + tstr + ', mesh = ' + nxstr + 'x' + nystr)
    savefig(pname)
    plt.close()

    # plot error in current solution (as a contour, and save to disk)
    t = tstep*dt;
    at = (1.0 - np.exp(-t*kprod))/kprod
    utrue = at*np.sin(np.pi*X)*np.sin(2.0*np.pi*Y);
    uerr = np.abs(utrue - results[tstep,:,:])
    plt.contourf(xspan,yspan,uerr,15, cmap=plt.cm.jet)
    plt.colorbar()
    plt.xlabel('x')
    plt.ylabel('y')
    plt.title('Error at output ' + tstr + ', mesh = ' + nxstr + 'x' + nystr)
    plt.savefig(cname)
    plt.close()




##### end of script #####
